"""Class for RESQML Horizon Interpretation organizational objects."""

from ._utils import (equivalent_extra_metadata, alias_for_attribute, extract_has_occurred_during,
                     equivalent_chrono_pairs, create_xml_has_occurred_during)

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns
from .genetic_boundary_feature import GeneticBoundaryFeature


class HorizonInterpretation(BaseResqpy):
    """Class for RESQML Horizon Interpretation organizational objects."""

    resqml_type = 'HorizonInterpretation'
    valid_domains = ('depth', 'time', 'mixed')
    valid_sequence_stratigraphy_surfaces = ('flooding', 'ravinement', 'maximum flooding', 'transgressive')
    valid_boundary_relations = ('conformable', 'unconformable below and above', 'unconformable above',
                                'unconformable below')

    def __init__(self,
                 parent_model,
                 root_node = None,
                 uuid = None,
                 title = None,
                 genetic_boundary_feature = None,
                 domain = 'depth',
                 boundary_relation_list = None,
                 sequence_stratigraphy_surface = None,
                 extra_metadata = None):
        """Initialises a horizon interpretation organisational object."""

        # note: will create a paired GeneticBoundaryFeature object when loading from xml (and possibly a Surface object)

        self.domain = domain
        self.genetic_boundary_feature = genetic_boundary_feature  # InterpretedFeature RESQML field, when not loading from xml
        self.feature_root = None if self.genetic_boundary_feature is None else self.genetic_boundary_feature.root
        if (not title) and self.genetic_boundary_feature is not None:
            title = self.genetic_boundary_feature.feature_name
        self.has_occurred_during = (None, None)
        self.boundary_relation_list = None if not boundary_relation_list else boundary_relation_list.copy()
        self.sequence_stratigraphy_surface = sequence_stratigraphy_surface
        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         extra_metadata = extra_metadata,
                         root_node = root_node)

    def _load_from_xml(self):
        self.domain = rqet.find_tag_text(self.root, 'Domain')
        interp_feature_ref_node = rqet.find_tag(self.root, 'InterpretedFeature')
        assert interp_feature_ref_node is not None
        self.feature_root = self.model.referenced_node(interp_feature_ref_node)
        if self.feature_root is not None:
            self.genetic_boundary_feature = GeneticBoundaryFeature(self.model,
                                                                   kind = 'horizon',
                                                                   uuid = self.feature_root.attrib['uuid'],
                                                                   feature_name = self.model.title_for_root(
                                                                       self.feature_root))
        self.has_occurred_during = extract_has_occurred_during(self.root)
        br_node_list = rqet.list_of_tag(self.root, 'BoundaryRelation')
        if br_node_list is not None and len(br_node_list) > 0:
            self.boundary_relation_list = []
            for br_node in br_node_list:
                self.boundary_relation_list.append(br_node.text)
        self.sequence_stratigraphy_surface = rqet.find_tag_text(self.root, 'SequenceStratigraphySurface')

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this interpretation is essentially the same as the other; otherwise False."""

        if other is None or not isinstance(other, HorizonInterpretation):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        if self.genetic_boundary_feature is not None:
            if not self.genetic_boundary_feature.is_equivalent(other.genetic_boundary_feature):
                return False
        elif other.genetic_boundary_feature is not None:
            return False
        if self.root is not None and other.root is not None:
            if rqet.citation_title_for_node(self.root) != rqet.citation_title_for_node(other.root):
                return False
        elif self.root is not None or other.root is not None:
            return False
        if (self.domain != other.domain or
                not equivalent_chrono_pairs(self.has_occurred_during, other.has_occurred_during) or
                self.sequence_stratigraphy_surface != other.sequence_stratigraphy_surface):
            return False
        if check_extra_metadata and not equivalent_extra_metadata(self, other):
            return False
        if not self.boundary_relation_list and not other.boundary_relation_list:
            return True
        if not self.boundary_relation_list or not other.boundary_relation_list:
            return False
        return set(self.boundary_relation_list) == set(other.boundary_relation_list)

    def create_xml(self,
                   genetic_boundary_feature_root = None,
                   add_as_part = True,
                   add_relationships = True,
                   originator = None,
                   title_suffix = None,
                   reuse = True):
        """Creates a horizon interpretation organisational xml node from a horizon interpretation object."""

        # note: related genetic boundary feature node should be created first and referenced here

        if not self.title:
            self.title = self.genetic_boundary_feature.feature_name
        if title_suffix:
            self.title += ' ' + title_suffix

        if reuse and self.try_reuse():
            return self.root
        hi = super().create_xml(add_as_part = False, originator = originator)

        if self.genetic_boundary_feature is not None:
            gbf_root = self.genetic_boundary_feature.root
            if gbf_root is not None:
                if genetic_boundary_feature_root is None:
                    genetic_boundary_feature_root = gbf_root
                else:
                    assert gbf_root is genetic_boundary_feature_root, 'genetic boundary feature mismatch'

        assert self.domain in self.valid_domains, 'illegal domain value for horizon interpretation'
        dom_node = rqet.SubElement(hi, ns['resqml2'] + 'Domain')
        dom_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
        dom_node.text = self.domain

        create_xml_has_occurred_during(self.model, hi, self.has_occurred_during)

        if self.boundary_relation_list is not None:
            for boundary_relation in self.boundary_relation_list:
                assert boundary_relation in self.valid_boundary_relations
                br_node = rqet.SubElement(hi, ns['resqml2'] + 'BoundaryRelation')
                br_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'BoundaryRelation')
                br_node.text = boundary_relation

        if self.sequence_stratigraphy_surface is not None:
            assert self.sequence_stratigraphy_surface in self.valid_sequence_stratigraphy_surfaces
            sss_node = rqet.SubElement(hi, ns['resqml2'] + 'SequenceStratigraphySurface')
            sss_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'SequenceStratigraphySurface')
            sss_node.text = self.sequence_stratigraphy_surface

        self.model.create_ref_node('InterpretedFeature',
                                   self.model.title_for_root(genetic_boundary_feature_root),
                                   genetic_boundary_feature_root.attrib['uuid'],
                                   content_type = 'obj_GeneticBoundaryFeature',
                                   root = hi)

        if add_as_part:
            self.model.add_part('obj_HorizonInterpretation', self.uuid, hi)
            if add_relationships:
                self.model.create_reciprocal_relationship(hi, 'destinationObject', genetic_boundary_feature_root,
                                                          'sourceObject')

        return hi



import pytest
import os
import numpy as np
import h5py

import resqpy.grid as grr
import resqpy.model as rq
import resqpy.property as rqp
import resqpy.olio.write_hdf5 as rqwh


def test_dtype_size(tmp_path):

    filenames = ['dtype_16', 'dtype_32', 'dtype_64']
    byte_sizes = [2, 4, 8]
    dtypes = [np.float16, np.float32, np.float64]
    hdf5_sizes = []

    extent_kji = (1000, 100, 100)
    a = np.random.random(extent_kji)

    for filename, dtype in zip(filenames, dtypes):
        epc = os.path.join(tmp_path, filename + '.epc')
        h5_file = epc[:-4] + '.h5'
        model = rq.new_model(epc)
        grid = grr.RegularGrid(model, extent_kji = extent_kji)
        grid.create_xml()
        pc = rqp.PropertyCollection()
        pc.set_support(support_uuid = grid.uuid, model = model)
        pc.add_cached_array_to_imported_list(cached_array = a,
                                             source_info = 'random',
                                             keyword = 'NTG',
                                             property_kind = 'net to gross ratio',
                                             indexable_element = 'cells',
                                             uom = 'm3/m3')
        pc.write_hdf5_for_imported_list(dtype = dtype)
        model.store_epc()
        model.h5_release()
        hdf5_sizes.append(os.path.getsize(h5_file))

    assert hdf5_sizes[0] < hdf5_sizes[1] < hdf5_sizes[2]
    for i, (byte_size, hdf5_size) in enumerate(zip(byte_sizes, hdf5_sizes)):
        array_size = byte_size * a.size
        # following may need to be modified if using hdf5 compression
        assert array_size < hdf5_size < array_size + 100000



# test point in polygon functions

import os

# import math as maths
import numpy as np

import resqpy.olio.point_inclusion as pip

# from numpy.testing import assert_array_almost_equal


def test_pip_cn_and_wn():
    # unit square polygon
    poly = np.array([(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0)])
    p_in = np.array([(0.00001, 0.00001), (0.00001, 0.99999), (0.99999, 0.00001), (0.99999, 0.99999)])
    p_out = np.array([(1.1, 0.1), (-0.1, 0.2), (0.5, 1.00001), (0.4, -0.0001), (1.00001, 1.00001), (1.00001, -0.00001)])
    for pip_fn in [pip.pip_cn, pip.pip_wn]:
        assert pip_fn((0.5, 0.5), poly)
        for p in p_in:
            assert pip_fn(p, poly)
        for p in p_out:
            assert not pip_fn(p, poly)
    assert np.all(pip.pip_array_cn(p_in, poly))
    assert not np.any(pip.pip_array_cn(p_out, poly))


def test_figure_of_eight():
    fig_8 = np.array([(-100.0, -200.0), (100.0, 200.0), (-100.0, 200.0), (100.0, -200.0)])
    p_in = np.array([(-99.0, -199.0), (0.0, -1.0), (99.0, 199.0), (0.0, 1.0), (49.9, -100.0)])
    p_out = np.array([(1000.0, -23.0), (1.0, 0.0), (-0.001, 0.0), (-50.1, 100.0)])
    for pip_fn in (pip.pip_cn, pip.pip_wn):
        for p in p_in:
            assert pip_fn(p, fig_8)
        for p in p_out:
            assert not pip_fn(p, fig_8)
    assert np.all(pip.pip_array_cn(p_in, fig_8))
    assert not np.any(pip.pip_array_cn(p_out, fig_8))


def test_points_in_polygon(tmp_path):
    # create an ascii file holding vertices of a polygon
    poly_file = os.path.join(tmp_path, 'diamond.txt')
    diamond = np.array([(0.0, 3.0, 0.0), (3.0, 6.0, -1.3), (6.0, 3.0, 12.5), (3.0, 0.0, 0.0)])
    with open(poly_file, 'w') as fp:
        for xyz in diamond:
            fp.write(f'{xyz[0]} {xyz[1]} {xyz[2]}\n')
        fp.write('999.0 999.0 999.0\n')
    # test some points with no multiplier applied to polygon geometry
    p_in = np.array([(3.0, 3.0), (0.1, 3.0), (1.55, 1.55), (4.49, 4.49), (5.99, 3.0), (3.1, 0.11)])
    p_out = np.array([(-3.0, -3.0), (2.0, 0.99), (4.51, 4.51), (6.01, 3.0)])
    assert np.all(pip.points_in_polygon(p_in[:, 0], p_in[:, 1], poly_file))
    assert not np.any(pip.points_in_polygon(p_out[:, 0], p_out[:, 1], poly_file))
    # test some points with multiplier applied to polygon geometry
    multiplier = 2.0 / 3.0
    p_in *= multiplier
    p_out *= multiplier
    assert np.all(pip.points_in_polygon(p_in[:, 0], p_in[:, 1], poly_file, poly_unit_multiplier = multiplier))
    assert not np.any(pip.points_in_polygon(p_out[:, 0], p_out[:, 1], poly_file, poly_unit_multiplier = multiplier))


def test_scan():
    kite = np.array([(0.1, 3.0, 0.0), (3.0, 4.9, 0.0), (5.9, 3.0, 0.0), (3.0, 0.1, 0.0)])
    origin = (-1.0, 0.0)
    ncol = 8
    nrow = 6
    expected = np.array([(0, 0, 0, 0, 0, 0, 0, 0), (0, 0, 0, 0, 1, 0, 0, 0), (0, 0, 0, 1, 1, 1, 0, 0),
                         (0, 0, 1, 1, 1, 1, 1, 0), (0, 0, 0, 1, 1, 1, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0)],
                        dtype = bool)
    assert np.all(pip.scan(origin, ncol, nrow, 1.0, 1.0, kite) == expected)



"""xml_namespaces.py: Module defining constant resqml xml namespaces."""

# resqml xml namespaces
# major revamp to support bespoke serialisation code, needed to emulate gSOAP, hence FESAPI

version = '15th February 2021'

namespace = {}

namespace['xsd'] = 'http://www.w3.org/2001/XMLSchema'
namespace['xsi'] = 'http://www.w3.org/2001/XMLSchema-instance'
# namespace['schema'] = 'http://www.w3.org/2001/XMLSchema-instance'

namespace['eml'] = 'http://www.energistics.org/energyml/data/commonv2'
namespace['eml20'] = 'http://www.energistics.org/energyml/data/commonv2'
namespace['eml23'] = 'http://www.energistics.org/energyml/data/commonv2'
namespace['resqml2'] = 'http://www.energistics.org/energyml/data/resqmlv2'
namespace['resqml'] = 'http://www.energistics.org/energyml/data/resqmlv2'
namespace['resqml20'] = 'http://www.energistics.org/energyml/data/resqmlv2'
namespace['rels_ext'] = 'http://schemas.energistics.org/package/2012/relationships/'

namespace['rels'] = 'http://schemas.openxmlformats.org/package/2006/relationships'
namespace['rels_md'] = 'http://schemas.openxmlformats.org/package/2006/relationships/metadata/'

namespace['content_types'] = 'http://schemas.openxmlformats.org/package/2006/content-types'
namespace['cp'] = 'http://schemas.openxmlformats.org/package/2006/metadata/core-properties'

namespace['dcterms'] = 'http://purl.org/dc/terms/'
namespace['dc'] = 'http://purl.org/dc/elements/1.1/'

curly_namespace = {}
for key, url in namespace.items():
    curly_namespace[key] = '{' + url + '}'

inverse_namespace = {}
for key, url in namespace.items():
    if url not in inverse_namespace:
        inverse_namespace[url] = key


def colon_namespace(url):
    """Returns the short form namespace for the url, complete with colon suffix."""

    if url[0] == '{':
        return inverse_namespace[url[1:-1]] + ':'
    return inverse_namespace[url] + ':'



# resqpy dataframe storage and retrieval test

import os

import numpy as np
import pandas as pd

import resqpy.model as rq
import resqpy.olio.dataframe as rqdf
import resqpy.time_series as rqts


def test_dataframe(tmp_path):

    epc = os.path.join(tmp_path, 'dataframe.epc')

    model = rq.new_model(epc_file = epc)

    np_df = np.array([[0.0, 0.0, 0.0], [1000.0, 0.0, 0.0], [1900.0, 100.0, 0.0], [2500.0, 500.0, 800.0],
                      [3200.0, 1500.0, 1600.0]])
    df_cols = ['COP', 'CWP', 'CWI']
    col_units = ['m3', 'ft3', 'm3']
    ts_list = ['2021-04-22', '2021-04-23', '2021-04-24', '2021-04-25', '2021-04-26']

    df = pd.DataFrame(np_df, columns = df_cols)

    vanilla = rqdf.DataFrame(model, df = df, title = 'vanilla')
    vanilla.write_hdf5_and_create_xml()

    ts = rqts.time_series_from_list(ts_list)
    ts.set_model(model)
    ts.create_xml()

    timetable = rqdf.TimeTable(model, df = df, title = 'time table', time_series = ts)
    timetable.write_hdf5_and_create_xml()

    propertied = rqdf.TimeTable(model, df = df, realization = 0, title = 'time table realisation', time_series = ts)
    propertied.write_hdf5_and_create_xml()

    united = rqdf.TimeTable(model,
                            df = df,
                            realization = 0,
                            title = 'united realisation',
                            uom_list = col_units,
                            time_series = ts)
    united.write_hdf5_and_create_xml()

    model.store_epc()
    model.h5_release()
    del model

    model = rq.Model(epc)
    df_roots = model.roots(obj_type = 'Grid2dRepresentation', extra = {'dataframe': 'true'})
    assert len(df_roots) == 4

    v_uuid = model.uuid(obj_type = 'Grid2dRepresentation', extra = {'dataframe': 'true'}, title = 'vanilla')
    dataframe = rqdf.DataFrame(model, uuid = v_uuid)
    assert dataframe.n_cols == 3
    assert dataframe.n_rows == 5
    assert all(dataframe.dataframe() == df)
    assert all(dataframe.dataframe().columns == df_cols)

    tt_uuid = model.uuid(obj_type = 'Grid2dRepresentation', extra = {'dataframe': 'true'}, title = 'time table')
    time_table = rqdf.TimeTable(model, uuid = tt_uuid)
    assert time_table.time_series().timestamps == ts_list
    assert all(time_table.dataframe() == df)

    u_uuid = model.uuid(obj_type = 'Grid2dRepresentation', extra = {'dataframe': 'true'}, title = 'united realisation')
    ur = rqdf.TimeTable(model, uuid = u_uuid)
    assert ur.uom_list == col_units
    assert ur.time_series().timestamps == ts_list

    # only dataframes with a realization number are stored as a property object
    assert len(model.parts(obj_type = 'ContinuousProperty')) == 2



# test crs creation and conversion from one crs to another

import os
import math as maths
import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

import resqpy.crs as rqc
import resqpy.model as rq
import resqpy.olio.uuid as bu


def test_crs(tmp_path):

    # create some coordinate reference systems
    model = rq.new_model(os.path.join(tmp_path, 'crs_test.epc'))
    crs_default = rqc.Crs(model)
    assert crs_default.null_transform
    crs_m = rqc.Crs(model, xy_units = 'm', z_units = 'm')
    crs_ft = rqc.Crs(model, xy_units = 'ft', z_units = 'ft')
    crs_mixed = rqc.Crs(model, xy_units = 'm', z_units = 'ft')
    crs_offset = rqc.Crs(model, xy_units = 'm', z_units = 'm', x_offset = 100.0, y_offset = -100.0, z_offset = -50.0)
    assert not crs_offset.null_transform
    crs_elevation = rqc.Crs(model, z_inc_down = False)
    crs_rotate = rqc.Crs(model, rotation = maths.pi / 2.0)
    crs_south = rqc.Crs(model, axis_order = 'southing westing')
    crs_time_s = rqc.Crs(model, xy_units = 'm', time_units = 's')
    crs_time_ms = rqc.Crs(model, xy_units = 'm', time_units = 'ms')
    for crs_time in [crs_time_s, crs_time_ms]:
        assert crs_time.resqml_type == 'LocalTime3dCrs'

    # check that distincitveness is recognised
    assert crs_default.is_equivalent(crs_m)
    assert not crs_m.is_equivalent(crs_ft)
    assert not crs_mixed.is_equivalent(crs_m)
    assert not crs_m.is_equivalent(crs_offset)
    assert not crs_m.is_equivalent(crs_elevation)
    assert not crs_m.is_equivalent(crs_rotate)
    assert not crs_m.is_equivalent(crs_south)
    assert not crs_time_s.is_equivalent(crs_time_ms)
    for depth_crs in [crs_default, crs_m, crs_ft, crs_mixed, crs_offset, crs_elevation, crs_rotate, crs_south]:
        assert depth_crs.resqml_type == 'LocalDepth3dCrs'
        assert not crs_time_s == depth_crs
        assert not crs_time_ms == depth_crs

    # check handedness
    assert not crs_m.is_right_handed_xy()
    assert not crs_m.is_right_handed_xyz()
    assert not crs_elevation.is_right_handed_xy()
    assert crs_elevation.is_right_handed_xyz()
    assert crs_south.is_right_handed_xy()
    assert crs_south.is_right_handed_xyz()

    # create some xml
    for crs in [
            crs_default, crs_m, crs_ft, crs_mixed, crs_offset, crs_elevation, crs_rotate, crs_south, crs_time_s,
            crs_time_ms
    ]:
        crs.create_xml()
    model.store_epc()
    # check re-use of equivalent crs'es
    assert bu.matching_uuids(crs_default.uuid, crs_m.uuid)

    # test conversion
    ft_to_m = 0.3048

    a = np.empty((10, 3))
    a[:, 0] = np.random.random(10) * 5.0e5
    a[:, 1] = np.random.random(10) * 10.0e5
    a[:, 2] = np.random.random(10) * 4.0e3

    b = a.copy()
    crs_m.convert_array_from(crs_default, a)
    assert np.max(np.abs(b - a)) < 1.0e-6
    a[:] = b
    crs_m.convert_array_to(crs_m, a)
    assert np.all(a == b)
    crs_ft.convert_array_from(crs_m, a)
    assert np.max(np.abs(b - a * ft_to_m)) < 1.0e-6
    crs_ft.convert_array_to(crs_m, a)
    assert np.max(np.abs(b - a)) < 1.0e-6
    a[:] = b
    crs_m.local_to_global_array(a)
    assert np.max(np.abs(b - a)) < 1.0e-6
    a[:] = b
    crs_offset.global_to_local_array(a)
    a[:, 0] += 100.0
    a[:, 1] -= 100.0
    a[:, 2] -= 50.0
    assert_array_almost_equal(a, b)

    # test single point conversion
    p = (456.78, 678.90, -1234.56)
    assert_array_almost_equal(p, crs_offset.global_to_local(crs_offset.local_to_global(p)))
    p_ft = crs_m.convert_to(crs_ft, np.array(p))
    assert_array_almost_equal(p, crs_m.convert_from(crs_ft, p_ft))

    # test time conversion
    pt = (123456.0, 234567.0, 1983.0)
    pt_s = np.array(crs_time_ms.convert_to(crs_time_s, pt))
    pt_s[2] *= 1000.0  # convert from seconds back to milliseconds
    assert_array_almost_equal(pt, pt_s)

    # todo: test rotation
    p = (234.00, 0.00, 5678.90)
    pr = crs_rotate.local_to_global(p)
    assert_array_almost_equal(pr, (0.00, 234.00, 5678.90))
    assert_array_almost_equal(crs_rotate.global_to_local(pr), p)


def test_crs_reuse():
    model = rq.Model(new_epc = True, create_basics = True)
    crs_a = rqc.Crs(model)
    crs_a.create_xml()
    crs_b = rqc.Crs(model)
    crs_b.create_xml()
    assert len(model.parts(obj_type = 'LocalDepth3dCrs')) == 1
    assert crs_a == crs_b
    assert bu.matching_uuids(crs_a.uuid, crs_b.uuid)
    crs_c = rqc.Crs(model, z_inc_down = False)
    crs_c.create_xml()
    assert len(model.parts(obj_type = 'LocalDepth3dCrs')) == 2
    assert crs_c != crs_a
    assert not bu.matching_uuids(crs_c.uuid, crs_a.uuid)
    crs_d = rqc.Crs(model, z_units = 'ft')
    crs_d.create_xml()
    assert len(model.parts(obj_type = 'LocalDepth3dCrs')) == 3
    crs_e = rqc.Crs(model, z_inc_down = False)
    crs_e.create_xml()
    assert len(model.uuids(obj_type = 'LocalDepth3dCrs')) == 3
    assert crs_e == crs_c
    assert bu.matching_uuids(crs_e.uuid, crs_c.uuid)
    crs_f = rqc.Crs(model)
    crs_f.create_xml(reuse = False)
    assert len(model.parts(obj_type = 'LocalDepth3dCrs')) == 4
    assert crs_f == crs_a
    assert not bu.matching_uuids(crs_f.uuid, crs_a.uuid)

    # todo: test parent epsg code equivalence



import numpy as np
import pandas as pd

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.organize as rqo
import resqpy.well


def test_load_from_xml(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    epc_path = model.epc_file
    elevation = 100
    # Create a measured depth datum
    datum = resqpy.well.MdDatum(parent_model = model,
                                crs_uuid = crs.uuid,
                                location = (0, 0, -elevation),
                                md_reference = 'kelly bushing')
    datum.create_xml()
    mds = np.array([300, 310, 330, 340])
    zs = mds - elevation
    well_name = 'JubJub'
    source_dataframe = pd.DataFrame({
        'MD': mds,
        'X': [1, 2, 3, 4],
        'Y': [1, 2, 3, 4],
        'Z': zs,
        'WELL': ['JubJub', 'JubJub', 'JubJub', 'JubJub']
    })

    # Create a trajectory from dataframe
    trajectory = resqpy.well.Trajectory(parent_model = model,
                                        data_frame = source_dataframe,
                                        well_name = well_name,
                                        md_datum = datum,
                                        length_uom = 'm')
    trajectory.write_hdf5()
    trajectory.create_xml()
    trajectory_uuid = trajectory.uuid

    # Create a wellbore marker frame
    wellbore_marker_frame = resqpy.well.WellboreMarkerFrame(parent_model = model,
                                                            trajectory_uuid = trajectory_uuid,
                                                            title = 'WMF1',
                                                            originator = 'Person1',
                                                            extra_metadata = {'target_reservoir': 'r1'})
    wellbore_marker_frame.create_xml()

    # Create several boundary features and interpretations
    horizon_feature_1 = rqo.GeneticBoundaryFeature(parent_model = model,
                                                   kind = 'horizon',
                                                   feature_name = 'horizon_feature_1')
    horizon_feature_1.create_xml()
    horizon_interp_1 = rqo.HorizonInterpretation(parent_model = model,
                                                 title = 'horizon_interp_1',
                                                 genetic_boundary_feature = horizon_feature_1,
                                                 sequence_stratigraphy_surface = 'flooding',
                                                 boundary_relation_list = ['conformable'])
    horizon_interp_1.create_xml()
    horizon_interp_uuid = horizon_interp_1.uuid

    fluid_contact_feature1 = rqo.FluidBoundaryFeature(parent_model = model,
                                                      kind = "gas oil contact",
                                                      feature_name = 'goc_1')
    fluid_contact_feature1.create_xml()

    # --------- Act ----------
    # Create a wellbore marker object
    wellbore_marker_1 = resqpy.well.WellboreMarker(parent_model = model,
                                                   parent_frame = wellbore_marker_frame,
                                                   marker_index = 0,
                                                   marker_type = 'horizon',
                                                   interpretation_uuid = horizon_interp_uuid,
                                                   title = 'Horizon1_marker',
                                                   extra_metadata = {'FormationName': 'Banoffee'})
    wellbore_marker_1_uuid = wellbore_marker_1.uuid
    wellbore_marker_2 = resqpy.well.WellboreMarker(parent_model = model,
                                                   parent_frame = wellbore_marker_frame,
                                                   marker_index = 1,
                                                   marker_type = 'gas oil contact',
                                                   title = 'GOC_marker')
    # Create xml for new wellbore markers
    wbm_node_1 = wellbore_marker_1.create_xml(parent_node = wellbore_marker_frame.root)
    wbm_node_2 = wellbore_marker_2.create_xml(parent_node = wellbore_marker_frame.root)

    # Load a new wellbore marker using a marker node
    wellbore_marker_3 = resqpy.well.WellboreMarker(parent_model = model,
                                                   parent_frame = wellbore_marker_frame,
                                                   marker_index = 0,
                                                   marker_node = wbm_node_1)

    # --------- Assert ----------
    assert rqet.find_tag_text(root = wbm_node_1, tag_name = 'GeologicBoundaryKind') == 'horizon'
    assert rqet.find_tag_text(root = wbm_node_2, tag_name = 'FluidContact') == 'gas oil contact'
    assert wellbore_marker_3 is not None
    assert wellbore_marker_3.title == 'Horizon1_marker'
    assert wellbore_marker_3.marker_type == 'horizon'
    assert bu.matching_uuids(wellbore_marker_3.interpretation_uuid, horizon_interp_uuid)
    assert wellbore_marker_3.extra_metadata == {'FormationName': 'Banoffee'}
    assert bu.matching_uuids(wellbore_marker_3.uuid, wellbore_marker_1_uuid)



"""_gcs_functions.py: Functions for working with grid connection sets."""

version = '20th October 2021'

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import os
import numpy as np

import resqpy.olio.xml_et as rqet
import resqpy.organize as rqo


def pinchout_connection_set(grid, skip_inactive = True, feature_name = 'pinchout'):
    """Returns a new GridConnectionSet representing non-standard K face connections across pinchouts.

    arguments:
       grid (grid.Grid): the grid for which a pinchout connection set is required
       skip_inactive (boolean, default True): if True, connections are not included where there is an inactive cell
          above or below the pinchout; if False, such connections are included
       feature_name (string, default 'pinchout'): the name to use as citation title in the feature and interpretation

    notes:
       this function does not write to hdf5, nor create xml for the new grid connection set;
       however, it does create one feature and a corresponding interpretation and creates xml for those
    """

    assert grid is not None

    po = grid.pinched_out()
    dead = grid.extract_inactive_mask() if skip_inactive else None

    cip_list = []  # cell index pair list

    for j in range(grid.nj):
        for i in range(grid.ni):
            ka = 0
            while True:
                while ka < grid.nk - 1 and po[ka, j, i]:
                    ka += 1
                while ka < grid.nk - 1 and not po[ka + 1, j, i]:
                    ka += 1
                if ka >= grid.nk - 1:
                    break
                # ka now in non-pinched out cell above pinchout
                if (skip_inactive and dead[ka, j, i]) or (grid.k_gaps and grid.k_gap_after_array[ka]):
                    ka += 1
                    continue
                kb = ka + 1
                while kb < grid.nk and po[kb, j, i]:
                    kb += 1
                if kb >= grid.nk:
                    break
                if skip_inactive and dead[kb, j, i]:
                    ka = kb + 1
                    continue
                # kb now beneath pinchout
                cip_list.append((grid.natural_cell_index((ka, j, i)), grid.natural_cell_index((kb, j, i))))
                ka = kb + 1

    log.debug(f'{len(cip_list)} pinchout connections found')

    pcs = _make_k_gcs_from_cip_list(grid, cip_list, feature_name)

    return pcs


def k_gap_connection_set(grid, skip_inactive = True, feature_name = 'k gap connection', tolerance = 0.001):
    """Returns a new GridConnectionSet representing K face connections where a K gap is zero thickness.

    arguments:
       grid (grid.Grid): the grid for which a K gap connection set is required
       skip_inactive (boolean, default True): if True, connections are not included where there is an inactive cell
          above or below the pinchout; if False, such connections are included
       feature_name (string, default 'pinchout'): the name to use as citation title in the feature and interpretation
       tolerance (float, default 0.001): the minimum vertical distance below which a K gap is deemed to be zero
          thickness; units are implicitly the z units of the coordinate reference system used by grid

    notes:
       this function does not write to hdf5, nor create xml for the new grid connection set;
       however, it does create one feature and a corresponding interpretation and creates xml for those;
       note that the entries in the connection set will be for logically K-neighbouring pairs of cells – such pairs
       are omitted from the standard transmissibilities due to the presence of the K gap layer
    """

    assert grid is not None
    if not grid.k_gaps:
        return None

    p = grid.points_ref(masked = False)
    dead = grid.extract_inactive_mask() if skip_inactive else None
    flip_z = (grid.k_direction_is_down != rqet.find_tag_bool(grid.crs_root, 'ZIncreasingDownward'))

    cip_list = []  # cell index pair list

    for k in range(grid.nk - 1):
        if not grid.k_gap_after_array[k]:
            continue
        k_gap_pillar_z = p[grid.k_raw_index_array[k + 1]][..., 2] - p[grid.k_raw_index_array[k] + 1][..., 2]
        if grid.has_split_coordinate_lines:
            pfc = grid.create_column_pillar_mapping()  # pillars for column
            k_gap_z = 0.25 * np.sum(k_gap_pillar_z[pfc], axis = (2, 3))  # resulting shape (nj, ni)
        else:
            k_gap_z = 0.25 * (k_gap_pillar_z[:-1, :-1] + k_gap_pillar_z[:-1, 1:] + k_gap_pillar_z[1:, :-1] +
                              k_gap_pillar_z[1:, 1:])  # shape (nj, ni)
        if flip_z:
            k_gap_z = -k_gap_z
        layer_mask = np.logical_and(np.logical_not(np.isnan(k_gap_z)), k_gap_z < tolerance)
        if skip_inactive:
            layer_mask = np.logical_and(layer_mask, np.logical_not(np.logical_or(dead[k], dead[k + 1])))
        # layer mask now boolean array of shape (nj, ni) set True where connection needed
        ji_list = np.stack(np.where(layer_mask)).T  # numpy array being list of [j, i] pairs
        for (j, i) in ji_list:
            cip_list.append((grid.natural_cell_index((k, j, i)), grid.natural_cell_index((k + 1, j, i))))

    log.debug(f'{len(cip_list)} k gap connections found')

    kgcs = _make_k_gcs_from_cip_list(grid, cip_list, feature_name)

    return kgcs


def add_connection_set_and_tmults(model, fault_incl, tmult_dict = None):
    """Add a grid connection set to a resqml model, based on a fault include file and a dictionary of fault:tmult pairs.

    Grid connection set added to resqml model, with extra_metadata on the fault interpretation containing the MULTFL values

    Args:
       model: resqml model object
       fault_incl: fullpath to fault include file or list of fullpaths to fault include files
       tmult_dict: dictionary of fault name/transmissibility multiplier pairs (must align with faults in include file).
          Optional, if blank values in the fault.include file will be used instead
    Returns:
       grid connection set uuid
    """

    from resqpy.fault._grid_connection_set import GridConnectionSet

    if isinstance(fault_incl, list):
        if len(fault_incl) > 1:
            # Making a concatenated version of the faultincl files
            # TODO: perhaps a better/more unique name and location could be used in future?
            temp_faults = os.path.join(os.path.dirname(model.epc_file), 'faults_combined_temp.txt')
            with open(temp_faults, 'w') as outfile:
                log.debug("combining multiple include files into one")
                for fname in fault_incl:
                    with open(fname) as infile:
                        for line in infile:
                            outfile.write(line)
        else:
            temp_faults = fault_incl[0]
    else:
        temp_faults = fault_incl

    log.info("Creating grid connection set")
    gcs = GridConnectionSet(parent_model = model,
                            ascii_load_format = 'nexus',
                            ascii_file = temp_faults,
                            create_organizing_objects_where_needed = True,
                            create_transmissibility_multiplier_property = True,
                            fault_tmult_dict = tmult_dict)
    gcs.write_hdf5(model.h5_file_name())
    gcs.create_xml(model.h5_uuid())

    model.store_epc()

    # clean up
    temp_combined_file = os.path.join(os.path.dirname(model.epc_file), 'faults_combined_temp.txt')
    if os.path.exists(temp_combined_file):
        os.remove(temp_combined_file)

    log.info("Grid connection set added")

    return gcs.uuid


# fault face table pandas dataframe functions
# these functions are for processing dataframes that have been read from (or to be written to) simulator ascii files


def zero_base_cell_indices_in_faces_df(faces, reverse = False):
    """Decrements all the cell indices in the fault face dataframe, in situ (or increments if reverse is True)."""

    if reverse:
        offset = 1
    else:
        offset = -1
    for col in ['i1', 'i2', 'j1', 'j2', 'k1', 'k2']:
        temp = faces[col] + offset
        faces[col] = temp


def standardize_face_indicator_in_faces_df(faces):
    """Sets face indicators to uppercase I, J or K, always with + or - following direction, in situ."""

    # todo: convert XYZ into IJK respectively?
    temp = faces['face'].copy().str.upper()
    temp[faces['face'].str.len() == 1] = faces['face'][faces['face'].str.len() == 1] + '+'
    for u in temp.unique():
        s = str(u)
        assert len(s) == 2, 'incorrect length to face string in fault dataframe: ' + s
        assert s[0] in 'IJK', 'unknown direction (axis) character in fault dataframe: ' + s
        assert s[1] in '+-', 'unknown face polarity character in fault dataframe: ' + s
    faces['face'] = temp


def remove_external_faces_from_faces_df(faces, extent_kji, remove_all_k_faces = False):
    """Returns a subset of the rows of faces dataframe, excluding rows on external faces."""

    # NB: assumes cell indices have been converted to zero based
    # NB: ignores grid column, ie. assumes extent_kji is applicable to all rows
    # NB: assumes single layer of cells is specified in the direction of the face
    filtered = []
    max_k0 = extent_kji[0] - 1
    max_j0 = extent_kji[1] - 1
    max_i0 = extent_kji[2] - 1
    for i in range(len(faces)):
        entry = faces.iloc[i]
        f = entry['face']
        if ((entry['i1'] <= 0 and f == 'I-') or (entry['j1'] <= 0 and f == 'J-') or (entry['k1'] <= 0 and f == 'K-') or
            (entry['i2'] >= max_i0 and f == 'I+') or (entry['j2'] >= max_j0 and f == 'J+') or
            (entry['k2'] >= max_k0 and f == 'K+')):
            continue
        if remove_all_k_faces and f[0] == 'K':
            continue
        filtered.append(i)
    return faces.loc[filtered]


def _make_k_gcs_from_cip_list(grid, cip_list, feature_name):
    # cip (cell index pair) list contains pairs of natural cell indices for which k connection is required
    # first of pair is layer above (lower k to be precise), second is below (higher k)
    # called by pinchout_connection_set() and k_gap_connection_set() functions

    from resqpy.fault._grid_connection_set import GridConnectionSet

    count = len(cip_list)

    if count == 0:
        return None

    pcs = GridConnectionSet(grid.model)
    pcs.grid_list = [grid]
    pcs.count = count
    pcs.grid_index_pairs = np.zeros((count, 2), dtype = int)
    pcs.cell_index_pairs = np.array(cip_list, dtype = int)
    pcs.face_index_pairs = np.zeros((count, 2), dtype = int)  # initialize to top faces
    pcs.face_index_pairs[:, 0] = 1  # bottom face of cells above pinchout

    pcs.feature_indices = np.zeros(count, dtype = int)  # could create seperate features by layer above or below?
    gbf = rqo.GeneticBoundaryFeature(grid.model, kind = 'horizon', feature_name = feature_name)
    gbf_root = gbf.create_xml()
    fi = rqo.HorizonInterpretation(grid.model, genetic_boundary_feature = gbf)
    fi_root = fi.create_xml(gbf_root, title_suffix = None)
    fi_uuid = rqet.uuid_for_part_root(fi_root)

    pcs.feature_list = [('obj_HorizonInterpretation', fi_uuid, str(feature_name))]

    return pcs



"""Code common to resqpy TimeSeries and GeologicTimeSeries classes."""

import logging

log = logging.getLogger(__name__)

import resqpy.time_series._functions as tsf
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns


class AnyTimeSeries(BaseResqpy):
    """Abstract class for a RESQML Time Series; use resqpy TimeSeries or GeologicTimeSeries.

    notes:
       this class has no direct initialisation method;
       call the any_time_series() function to generate the appropriate derived class object for a given uuid;
       the resqpy code differentiates between time series on a human timeframe, where the timestamps are
       used without a year offset, and those on a geologic timeframe, where the timestamps are ignored and
       only the year offset is significant
    """

    resqml_type = 'TimeSeries'

    def _load_from_xml(self):
        time_series_root = self.root
        assert time_series_root is not None
        # for human timeframe series, timestamps is an ordered list of timestamp strings in resqml/iso format
        # for geological timeframe series, timestamps is an ordered list of ints being the year offsets from present
        self.timestamps = []
        for child in rqet.list_of_tag(time_series_root, 'Time'):
            dt_text = rqet.find_tag_text(child, 'DateTime')
            assert dt_text, 'missing DateTime field in xml for time series'
            year_offset = rqet.find_tag_int(child, 'YearOffset')
            if year_offset:
                assert self.timeframe == 'geologic'
                self.timestamps.append(year_offset)  # todo: trim and check timestamp
            else:
                assert self.timeframe == 'human'
                self.timestamps.append(dt_text)  # todo: trim and check timestamp
            self.timestamps.sort()

    def is_equivalent(self, other_ts):
        """Performs partial equivalence check on two time series (derived class methods finish the work)."""

        if other_ts is None:
            return False
        if self is other_ts:
            return True
        if bu.matching_uuids(self.uuid, other_ts.uuid):
            return True
        if self.number_of_timestamps() != other_ts.number_of_timestamps():
            return False
        if self.timeframe != other_ts.timeframe:
            return False
        return None

    def set_model(self, parent_model):
        """Associate the time series with a resqml model (does not create xml or write hdf5 data)."""
        self.model = parent_model

    def number_of_timestamps(self):
        """Returns the number of timestamps in the series.

        :meta common:
        """
        return len(self.timestamps)

    def timestamp(self, index, as_string = True):
        """Returns an individual timestamp, indexed by its position in the series.

        arguments:
           index (int): the time index for which the timestamp is required
           as_string (boolean, default True): if True and this is series is on a geologic timeframe, the return
              value is a string, otherwise an int; for human timeframe series, this argument has no effect

        returns:
           string or int being the selected timestamp

        notes:
           index may be negative in which case it is taken to be relative to the end of the series
           with the last timestamp being referenced by an index of -1;
           the string form of a geologic timestamp is a positive number in millions of years,
           with the suffix Ma

        :meta common:
        """

        # individual timestamp is in iso format with a Z appended, eg: 2019-08-23T14:30:00Z
        if not -len(self.timestamps) <= index < len(self.timestamps):
            return None
        stamp = self.timestamps[index]
        if as_string and isinstance(stamp, int):
            stamp = tsf.geologic_time_str(stamp)
        return stamp

    def iter_timestamps(self, as_string = True):
        """Iterator over timestamps.

        :meta common:
        """
        for index in range(len(self.timestamps)):
            yield self.timestamp(index, as_string = as_string)

    def last_timestamp(self, as_string = True):
        """Returns the last timestamp in the series.

        :meta common:
        """
        return self.timestamp(-1, as_string = as_string) if self.timestamps else None

    def index_for_timestamp(self, timestamp):
        """Returns the index for a given timestamp.

        note:
           this method uses a simplistic string or int comparison;
           if the timestamp is not found, None is returned

        :meta common:
        """
        as_string = not isinstance(timestamp, int)
        for index in range(len(self.timestamps)):
            if self.timestamp(index, as_string = as_string) == timestamp:
                return index
        return None

    def create_xml(self, add_as_part = True, title = None, originator = None, reuse = True):
        """Create a RESQML time series xml node from a TimeSeries or GeologicTimeSeries object, optionally add as part.

        arguments:
           add_as_part (boolean, default True): if True, the newly created xml node is added as a part
              in the model
           title (string): used as the citation Title text for the new time series node
           originator (string, optional): the name of the human being who created the time series object;
              default is to use the login name

        returns:
           the newly created time series xml node

        :meta common:
        """

        if reuse and self.try_reuse():
            return self.root  # check for reusable (equivalent) object

        if self.extra_metadata is None:
            self.extra_metadata = {}
        self.extra_metadata['timeframe'] = self.timeframe

        ts_node = super().create_xml(add_as_part = False, title = title, originator = originator)

        for index in range(self.number_of_timestamps()):
            time_node = rqet.SubElement(ts_node, ns['resqml2'] + 'Time')
            time_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Timestamp')
            time_node.text = rqet.null_xml_text
            dt_node = rqet.SubElement(time_node, ns['resqml2'] + 'DateTime')
            dt_node.set(ns['xsi'] + 'type', ns['xsd'] + 'dateTime')
            if self.timeframe == 'geologic':
                assert isinstance(self.timestamps[index], int)
                dt_node.text = '0000-01-01T00:00:00Z'
                yo_node = rqet.SubElement(time_node, ns['resqml2'] + 'YearOffset')
                yo_node.set(ns['xsi'] + 'type', ns['xsd'] + 'long')
                yo_node.text = str(self.timestamps[index])
            else:
                dt_node.text = self.timestamp(index)

        if add_as_part:
            self.model.add_part('obj_TimeSeries', self.uuid, ts_node)

        return ts_node

    def create_time_index(self, time_index, root = None, check_valid_index = True):
        """Create a time index node, including time series reference, optionally add to root.

        arguments:
           time_index (int): non-negative integer index into the time series, identifying the datetime
              being referenced by the new node
           root (optional): if present, the newly created time index xml node is added
              as a child to this node
           check_valid_index (boolean, default True): if True, an assertion error is raised
              if the time index is out of range for this time series

        returns:
           the newly created time index xml node

        note:
           a time index node is typically part of a recurrent property object; it identifies
           the datetime relevant to the property array (or other data) by indexing into a time series
           object
        """

        assert self.uuid is not None
        if check_valid_index:
            assert 0 <= time_index < len(self.timestamps), 'time index out of range for time series'

        t_node = rqet.Element(ns['resqml2'] + 'TimeIndex')
        t_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'TimeIndex')
        t_node.text = rqet.null_xml_text

        index_node = rqet.SubElement(t_node, ns['resqml2'] + 'Index')
        index_node.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
        index_node.text = str(time_index)

        self.model.create_time_series_ref(self.uuid, root = t_node)

        if root is not None:
            root.append(t_node)
        return t_node



import os
import math as maths
import numpy as np
from numpy.testing import assert_array_almost_equal

import resqpy.lines
import resqpy.model as rq
import resqpy.organize


def test_lines(example_model_and_crs):

    # Set up a Polyline
    title = 'Nazca'
    model, crs = example_model_and_crs
    line = resqpy.lines.Polyline(parent_model = model,
                                 title = title,
                                 set_crs = crs.uuid,
                                 set_crsroot = crs.root,
                                 set_bool = True,
                                 set_coord = np.array([[0, 0, 0], [1, 1, 1]]))
    line.write_hdf5()
    line.create_xml()

    # Add a interpretation
    assert line.rep_int_root is None
    line.create_interpretation_and_feature(kind = 'fault')
    assert line.rep_int_root is not None

    # Check fault can be loaded in again
    model.store_epc()
    model = rq.Model(epc_file = model.epc_file)
    reload = resqpy.lines.Polyline(parent_model = model, uuid = line.uuid)
    assert reload.citation_title == title

    fault_interp = resqpy.organize.FaultInterpretation(model, uuid = line.rep_int_uuid)
    fault_feature = resqpy.organize.TectonicBoundaryFeature(model, uuid = fault_interp.feature_uuid)

    # Check title matches expected title
    assert fault_feature.feature_name == title


def test_lineset(example_model_and_crs, tmp_path):

    # Set up a PolylineSet
    title = 'Nazcas'
    model, crs = example_model_and_crs
    line1 = resqpy.lines.Polyline(parent_model = model,
                                  title = title,
                                  set_crs = crs.uuid,
                                  set_crsroot = crs.root,
                                  set_bool = True,
                                  set_coord = np.array([[0, 0, 0], [1, 1, 1]], dtype = float))

    line2 = resqpy.lines.Polyline(parent_model = model,
                                  title = title,
                                  set_crs = crs.uuid,
                                  set_crsroot = crs.root,
                                  set_bool = True,
                                  set_coord = np.array([[2, 2, 2], [3, 3, 3]], dtype = float))

    lines = resqpy.lines.PolylineSet(parent_model = model, title = title, polylines = [line1, line2])

    lines.write_hdf5()
    lines.create_xml()

    # Check lines can be loaded in again
    model.store_epc()
    model = rq.Model(epc_file = model.epc_file)
    reload = resqpy.lines.PolylineSet(parent_model = model, uuid = lines.uuid)
    assert len(reload.polys) == 2,  \
        f'Expected two polylines in the polylineset, found {len(reload.polys)}'
    assert (reload.count_perpol == [2, 2]).all(),  \
        f'Expected count per polyline to be [2,2], found {reload.count_perpol}'
    pl_list = reload.convert_to_polylines()
    assert len(pl_list) == 2
    pl_set_2 = resqpy.lines.PolylineSet(model)
    pl_set_2.combine_polylines(pl_list)
    assert len(pl_set_2.polys) == 2,  \
        f'Expected two polylines in the polylineset, found {len(pl_set_2.polys)}'
    assert (pl_set_2.count_perpol == [2, 2]).all(),  \
        f'Expected count per polyline to be [2,2], found {pl_set_2.count_perpol}'
    irap_file = os.path.join(tmp_path, 'test_irap.dat')
    pl_set_2.convert_to_irap(irap_file)
    charisma_file = os.path.join(tmp_path, 'test_charisma.dat')
    pl_set_2.convert_to_charisma(charisma_file)
    pl_set_3 = resqpy.lines.PolylineSet(model, charisma_file = charisma_file)
    assert pl_set_3 is not None and len(pl_set_3.polys) == 2
    pl_set_4 = resqpy.lines.PolylineSet(model, irap_file = irap_file)
    assert pl_set_4 is not None and len(pl_set_4.polys) == 2


def test_charisma(example_model_and_crs, test_data_path):
    # Set up a PolylineSet
    model, crs = example_model_and_crs
    charisma_file = test_data_path / "Charisma_example.txt"
    lines = resqpy.lines.PolylineSet(parent_model = model, charisma_file = str(charisma_file))
    lines.write_hdf5()
    lines.create_xml()

    model.store_epc()
    model = rq.Model(epc_file = model.epc_file)
    reload = resqpy.lines.PolylineSet(parent_model = model, uuid = lines.uuid)

    assert reload.title == 'Charisma_example'
    assert (reload.count_perpol == [4, 5, 4, 5, 5]).all(),  \
        f"Expected count per polyline to be [4,5,4,5,5], found {reload.count_perpol}"
    assert len(reload.coordinates) == 23,  \
        f"Expected length of coordinates to be 23, found {len(reload.coordinates)}"


def test_irap(example_model_and_crs, test_data_path):
    # Set up a PolylineSet
    model, crs = example_model_and_crs
    irap_file = test_data_path / "IRAP_example.txt"
    lines = resqpy.lines.PolylineSet(parent_model = model, irap_file = str(irap_file))
    lines.write_hdf5()
    lines.create_xml()

    model.store_epc()
    model = rq.Model(epc_file = model.epc_file)
    reload = resqpy.lines.PolylineSet(parent_model = model, uuid = lines.uuid)

    assert reload.title == 'IRAP_example'
    assert (reload.count_perpol == [15]).all(),  \
        f"Expected count per polyline to be [15], found {reload.count_perpol}"
    assert len(reload.coordinates) == 15,  \
        f"Expected length of coordinates to be 15, found {len(reload.coordinates)}"


def test_is_clockwise(example_model_and_crs):
    # Set up a Polyline
    title = 'diamond'
    model, crs = example_model_and_crs
    line = resqpy.lines.Polyline(
        parent_model = model,
        title = title,
        set_crs = crs.uuid,
        set_bool = True,
        set_coord = np.array([
            (3.0, 2.0, 4.3),  # z values are ignored for clockwise test
            (2.0, 1.0, -13.5),
            (1.0, 2.5, 7.8),
            (2.1, 3.9, -2.0)
        ]))
    assert line is not None
    for trust in [False, True]:
        assert line.is_clockwise(trust_metadata = trust)
    # reverse the order of the coordinates
    line.coordinates = np.flip(line.coordinates, axis = 0)
    assert line.is_clockwise(trust_metadata = True)
    assert not line.is_clockwise(trust_metadata = False)  # should reset metadata
    assert not line.is_clockwise(trust_metadata = True)


def test_is_convex(example_model_and_crs):
    # Set up a Polyline
    title = 'pentagon'
    model, crs = example_model_and_crs
    line = resqpy.lines.Polyline(
        parent_model = model,
        title = title,
        set_crs = crs.uuid,
        set_bool = True,
        set_coord = np.array([
            (3.0, 2.0, 4.3),  # z values are ignored for convexity
            (2.0, 1.0, -13.5),
            (1.0, 2.5, 7.8),
            (1.5, 3.0, -2.0),
            (2.5, 3.0, 23.4)
        ]))
    assert line is not None
    for trust in [False, True]:
        assert line.is_convex(trust_metadata = trust)
    # adjust one point to make shape concave
    line.coordinates[4, 1] = 1.9
    assert line.is_convex(trust_metadata = True)
    assert not line.is_convex(trust_metadata = False)  # should reset metadata
    assert not line.is_convex(trust_metadata = True)


def test_from_scaled_polyline(example_model_and_crs):
    # Set up an original Polyline
    title = 'rectangle'
    model, crs = example_model_and_crs
    line = resqpy.lines.Polyline(parent_model = model,
                                 title = title,
                                 set_crs = crs.uuid,
                                 set_bool = True,
                                 set_coord = np.array([(0.0, 0.0, 10.0), (0.0, 3.0, 10.0), (12.0, 3.0, 10.0),
                                                       (12.0, 0.0, 10.0)]))
    assert line is not None

    # test scaling up to a larger polyline
    bigger = resqpy.lines.Polyline.from_scaled_polyline(line, 1.5, originator = 'testing')
    expected = np.array([(-3.0, -0.75, 10.0), (-3.0, 3.75, 10.0), (15.0, 3.75, 10.0), (15.0, -0.75, 10.0)])
    assert bigger.isclosed
    assert bigger.is_convex()
    assert bigger.title == 'rectangle'
    assert bigger.originator == 'testing'
    assert_array_almost_equal(bigger.coordinates, expected)

    # test scaling to a smaller polyline
    smaller = resqpy.lines.Polyline.from_scaled_polyline(line, 2.0 / 3.0, title = 'small')
    expected = np.array([(2.0, 0.5, 10.0), (2.0, 2.5, 10.0), (10.0, 2.5, 10.0), (10.0, 0.5, 10.0)])
    assert smaller.title == 'small'
    assert_array_almost_equal(smaller.coordinates, expected)


def test_point_is_inside_and_balanced_centre_and_segment_normal(example_model_and_crs):
    # Set up an original Polyline
    title = 'sigma'
    model, crs = example_model_and_crs
    line = resqpy.lines.Polyline(parent_model = model,
                                 title = title,
                                 set_crs = crs.uuid,
                                 set_bool = True,
                                 set_coord = np.array([(0.0, 0.0, 10.0), (0.0, 3.0, 10.0), (4.0, 3.0, 10.0),
                                                       (2.0, 1.5, 10.0), (4.0, 0.0, 10.0)]))
    assert line is not None
    assert not line.is_convex()
    assert line.is_clockwise(trust_metadata = False)
    centre = line.balanced_centre(in_xy = True)
    for mode in ['crossing', 'winding']:
        assert line.point_is_inside_xy((1.0, 1.0, 1.0), mode = mode)
        assert not line.point_is_inside_xy((2.5, 1.4, 1.0), mode = mode)
        assert line.point_is_inside_xy(centre, mode = mode)
    # check a couple of normal vectors
    assert_array_almost_equal(line.segment_normal(1), (0.0, 1.0, 0.0))
    assert_array_almost_equal(line.segment_normal(4), (0.0, -1.0, 0.0))


def test_segment_methods(example_model_and_crs):
    # Set up an original Polyline
    model, crs = example_model_and_crs
    line = __zig_zag(model, crs)
    coords = line.coordinates
    assert maths.isclose(line.segment_length(0), maths.sqrt(27.0))
    assert maths.isclose(line.segment_length(0, in_xy = True), maths.sqrt(18.0))
    assert_array_almost_equal(line.segment_midpoint(1), (3.0, 8.0, 13.5))
    for in_xy in [False, True]:
        d = 2 if in_xy else 3
        length = 0.0
        for seg in range(len(coords) - 1):
            length += maths.sqrt(
                np.sum((coords[seg + 1, :d] - coords[seg, :d]) * (coords[seg + 1, :d] - coords[seg, :d])))
        assert maths.isclose(line.full_length(in_xy = in_xy), length)
    # check interpolation method at end points
    assert_array_almost_equal(line.interpolated_point(0.0), coords[0])
    assert_array_almost_equal(line.interpolated_point(1.0), coords[-1])
    # create a simple 3 segment polyline
    coords = np.array([(0.0, 0.0, 0.0), (5.0, 0.0, 0.0), (10.0, 5.0, 0.0), (10.0, 10.0, 0.0)])
    line = resqpy.lines.Polyline(parent_model = model,
                                 title = 'angled',
                                 set_crs = crs.uuid,
                                 set_bool = False,
                                 set_coord = coords)
    # check midpoint
    assert_array_almost_equal(line.interpolated_point(0.5), (7.5, 2.5, 0.0))
    # create some equidistant points along the line
    ep = line.equidistant_points(7)
    # check some of the equidistant points
    assert_array_almost_equal(ep[0], coords[0])
    assert maths.isclose(ep[1, 1], 0.0)
    assert_array_almost_equal(ep[3], (7.5, 2.5, 0.0))
    assert maths.isclose(ep[5, 0], 10.0)
    assert_array_almost_equal(ep[-1], coords[-1])
    # check line segment intersection method
    for half_seg in [False, True]:
        seg, x, y = line.first_line_intersection(5.0, 5.0, 10.0, 0.0, half_segment = half_seg)
        assert seg == 1
        assert maths.isclose(x, 7.5) and maths.isclose(y, 2.5)
        seg, x, y = line.first_line_intersection(5.0, 5.0, 0.0, 10.0, half_segment = half_seg)
        assert seg is None and x is None and y is None


def test_area(example_model_and_crs):
    # create an octagonal polyline
    model, crs = example_model_and_crs
    line = __octagon(model, crs)
    assert maths.isclose(line.area(), 14.0)


def test_splined_and_tangent_vectors(example_model_and_crs):
    model, crs = example_model_and_crs
    line = __zig_zag(model, crs)
    s_line = line.splined(title = 'splined zig zag')
    assert s_line is not None
    tans = s_line.tangent_vectors()
    assert tans is not None
    assert len(tans) == len(s_line.coordinates)


def test_normalised_and_denormalised(example_model_and_crs):
    model, crs = example_model_and_crs
    line = __octagon(model, crs)
    for mode in ['square', 'perimeter', 'circle']:
        xn, yn = line.normalised_xy(3.0, 3.0, mode = mode)
        if mode == 'square':
            assert maths.isclose(xn, yn)
            assert 0.0 < xn < 0.5
        elif mode == 'perimeter':
            assert maths.isclose(yn, 0.0) or maths.isclose(yn, 1.0)
        x, y = line.denormalised_xy(xn, yn, mode = mode)
        assert maths.isclose(x, y)
        assert maths.isclose(x, 3.0)
        if mode != 'square':
            xn, yn = line.normalised_xy(4.25, 5.13, mode = mode)
            x, y = line.denormalised_xy(xn, yn, mode = mode)
            assert maths.isclose(x, 4.25) and maths.isclose(y, 5.13)


def test_poly_index_containing_point_in_xy(example_model_and_crs):
    model, crs = example_model_and_crs
    octagon = __octagon(model, crs)
    square = __square(model, crs)
    pl_set = resqpy.lines.PolylineSet(model, polylines = [octagon, square])
    for mode in ['crossing', 'winding']:
        assert pl_set.poly_index_containing_point_in_xy((5.8, 5.8), mode = mode) is None
        assert pl_set.poly_index_containing_point_in_xy((5.1, 5.1), mode = mode) == 0
        assert pl_set.poly_index_containing_point_in_xy((6.2, 4.8, 0.0), mode = mode) == 1


def __zig_zag(model, crs):
    title = 'zig_zag'
    coords = np.array([(4.0, 5.0, 10.0), (1.0, 8.0, 13.0), (5.0, 8.0, 14.0), (1.0, 12.0, 10.0), (5.0, 12.0, 10.0)])
    line = resqpy.lines.Polyline(parent_model = model,
                                 title = title,
                                 set_crs = crs.uuid,
                                 set_bool = False,
                                 set_coord = coords)
    return line


def __octagon(model, crs):
    title = 'octagon'
    coords = np.array([(2.5, 2.5, 0.0), (2.0, 3.0, 0.0), (2.0, 5.0, 0.0), (3.0, 6.0, 0.0), (5.0, 6.0, 0.0),
                       (6.0, 5.0, 0.0), (6.0, 3.0, 0.0), (5.0, 2.0, 0.0), (3.0, 2.0, 0.0)])
    line = resqpy.lines.Polyline(parent_model = model,
                                 title = title,
                                 set_crs = crs.uuid,
                                 set_bool = True,
                                 set_coord = coords)
    return line


def __square(model, crs):
    title = 'square'
    coords = np.array([(6.0, 3.0, 10.0), (6.0, 5.0, 10.0), (8.0, 5.0, 10.0), (8.0, 3.0, 10.0)])
    line = resqpy.lines.Polyline(parent_model = model,
                                 title = title,
                                 set_crs = crs.uuid,
                                 set_bool = True,
                                 set_coord = coords)
    return line



"""rq_print.py: simple print functionality for summarizing resqml objects."""

version = '1st July 2021'

import logging

log = logging.getLogger(__name__)

import h5py
import numpy as np

# todo: redistribute functions to main object class modules and remove circular imports
import resqpy.crs as rqc
import resqpy.fault as rqf
import resqpy.grid as grr
import resqpy.olio.class_dict as rcd
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.property as rprop
import resqpy.time_series as rts
import resqpy.well as rqw


def pl(n, use_e = False):
    """Returns a string suffix for indicating plurality of n."""
    if n == 1:
        return ''
    if use_e:
        return 'es'
    return 's'


def required(value):
    """Returns value as str, '(missing)', or '(empty)'."""
    if value is None:
        return '(missing)'
    text = str(value)
    if len(text):
        return text
    return '(empty)'


def optional(value):
    """Returns value as str, '(not present)', or '(empty)'."""
    if value is None:
        return '(not present)'
    text = str(value)
    if len(text):
        return text
    return '(empty)'


def format_xyz(xyz):
    """Returns a formatted string representing triple float xyz values."""
    x, y, z = xyz
    if x is None or y is None or z is None:
        return '(invalid)'
    return '(x: {0:5.3f}, y: {1:5.3f}, z: {2:5.3f})'.format(x, y, z)


def point_3d(node, flavour):
    """Extracts and returns a triple float xyz from a RESQML Point3d xml node."""
    p3d_node = rqet.find_tag(node, flavour)
    if p3d_node is None:
        return '(missing)'
    x = rqet.find_tag_float(p3d_node, 'Coordinate1')
    y = rqet.find_tag_float(p3d_node, 'Coordinate2')
    z = rqet.find_tag_float(p3d_node, 'Coordinate3')
    return (x, y, z)


def print_citation(node):
    """Prints information from a citation xml tree."""

    if node is None:
        return
    c_node = rqet.find_tag(node, 'Citation')
    if c_node is None:
        return
    print('citation:')
    print('   title:   ' + str(rqet.node_text(rqet.find_tag(c_node, 'Title'))))
    print('   user:    ' + str(rqet.node_text(rqet.find_tag(c_node, 'Originator'))))
    print('   created: ' + str(rqet.node_text(rqet.find_tag(c_node, 'Creation'))))
    print('   format:  ' + str(rqet.node_text(rqet.find_tag(c_node, 'Format'))))
    description = rqet.find_tag_text(c_node, 'Description')
    if description:
        print('   description: ' + str(description))


def print_LocalDepth3dCrs(model, node, z_is_time = False):
    """Prints information from a coordinate reference system xml tree."""

    print('units xy: ' + str(rqet.length_units_from_node(rqet.find_tag(node, 'ProjectedUom'))))
    print('units z:  ' + str(rqet.length_units_from_node(rqet.find_tag(node, 'VerticalUom'))))  # applicable to time?
    xy_axes = rqet.node_text(rqet.find_tag(node, 'ProjectedAxisOrder'))
    print('x y axes: ' + str(xy_axes))
    z_inc_down = rqet.bool_from_text(rqet.node_text(rqet.find_tag(
        node, 'ZIncreasingDownward')))  # todo: check applicability to Time3dCrs?
    if z_inc_down is None:
        z_dir = 'unknown'
    elif z_inc_down:
        z_dir = 'downwards'
    else:
        z_dir = 'upwards'
    print('z increases: ' + str(z_dir))
    print('(xyz handedness: ' + str(rqet.xyz_handedness(xy_axes, z_inc_down) + ')'))
    print('x offset: ' + str(rqet.node_text(rqet.find_tag(node, 'XOffset'))))
    print('y offset: ' + str(rqet.node_text(rqet.find_tag(node, 'YOffset'))))
    print('z offset: ' + str(rqet.node_text(rqet.find_tag(node, 'ZOffset'))))  # todo: check applicability to Time3dCrs?
    print('rotation: ' + str(rqet.node_text(rqet.find_tag(node, 'ArealRotation'))))
    parent_xy_crs_node = rqet.find_tag(node, 'ProjectedCrs')
    parent_xy_crs_node_type = rqet.node_type(parent_xy_crs_node)
    if parent_xy_crs_node_type == 'ProjectedCrsEpsgCode':
        print('parent xy crs: EPSG code ' + str(rqet.node_text(rqet.find_tag(parent_xy_crs_node, 'EpsgCode'))))
    elif parent_xy_crs_node_type == 'ProjectedUnknownCrs':
        print('parent xy crs: unknown')
    parent_z_crs_node = rqet.find_tag(node, 'VerticalCrs')  # todo: check applicability to Time3dCrs?
    parent_z_crs_node_type = rqet.node_type(parent_z_crs_node)
    if parent_z_crs_node_type == 'VerticalCrsEpsgCode':
        print('parent z crs: EPSG code ' + str(rqet.node_text(rqet.find_tag(parent_z_crs_node, 'EpsgCode'))))
    elif parent_z_crs_node_type == 'VerticalUnknownCrs':
        print('parent z crs: unknown')
    # todo: something with parent xy Crs & z Crs other than EPSG or Unknown


def print_LocalTime3dCrs(model, node):
    """Prints information from a time based (seismic) coordinate reference system xml tree."""

    # very similar to obj_LocalDepth3dCrs, with added TimeUom
    print_LocalDepth3dCrs(model, node, z_is_time = True)
    print('time units: ' + str(rqet.time_units_from_node(rqet.find_tag(node, 'TimeUom'))))


def print_IjkGridRepresentation(model, node, detail_level = 0, grid = None):
    """Prints information about an IJK Grid object, mostly from the xml tree."""

    print('ni: ' + str(rqet.node_text(rqet.find_tag(node, 'Ni'))))
    print('nj: ' + str(rqet.node_text(rqet.find_tag(node, 'Nj'))))
    print('nk: ' + str(rqet.node_text(rqet.find_tag(node, 'Nk'))))
    k_gaps_node = rqet.find_tag(node, 'KGaps')
    if k_gaps_node is None:
        print('grid does not have k gaps')
    else:
        k_gaps_count = rqet.find_tag_int(k_gaps_node, 'Count')
        print('number of k gaps: ' + str(k_gaps_count))
    pw_node = rqet.find_tag(node, 'ParentWindow')
    if pw_node is not None:
        parent_uuid = rqet.find_nested_tags_text(pw_node, ['ParentGrid', 'UUID'])
        print('parent window present (grid is a local grid); parent uuid: ' + str(parent_uuid))
    geom_node = rqet.find_tag(node, 'Geometry')
    if geom_node is None:
        if pw_node is not None:
            print('geometry is not present (to be derived from parent grid)')
        else:
            print('geometry is missing!')
        return
    print('k increases: ' + str(rqet.find_tag_text(geom_node, 'KDirection')))
    print('ijk handedness: ' + str(rqet.ijk_handedness(geom_node)))
    print('pillar shape: ' + str(rqet.find_tag_text(geom_node, 'PillarShape')))
    if not detail_level:
        return
    # create a Grid object and interrogate for extra details
    if grid is None:
        grid = grr.Grid(model, grid_root = node)
    cell_count = grid.cell_count()
    print('(number of cells: ' + str(cell_count) + ')')
    if grid.has_split_coordinate_lines:
        print('grid has split pillars (faults)')
        split_root = grid.resolve_geometry_child('SplitCoordinateLines')
        assert split_root is not None
        print('number of split pillars: ' + str(rqet.node_text(rqet.find_tag(split_root, 'Count'))))
    else:
        print('grid does not have split pillars (grid is unfaulted)')
    pillar_is_def_root = grid.resolve_geometry_child('PillarGeometryIsDefined')
    cells_all_defined = True
    if pillar_is_def_root is None:
        print('grid does not include boolean array indicating which pillar geometries are defined!')
    else:
        pillar_is_def_type = rqet.node_type(pillar_is_def_root)
        if pillar_is_def_type == 'BooleanConstantArray':
            pillars_all_defined = grid.pillar_geometry_is_defined()
            assert pillars_all_defined is not None
            if pillars_all_defined:
                print('geometry is defined for all pillars')
            else:
                print('geometry is not defined for any pillar!')
            cells_all_defined = pillars_all_defined
        else:
            assert pillar_is_def_type == 'BooleanHdf5Array'
            if detail_level > 1:
                if grid.geometry_defined_for_all_pillars():
                    print('geometry is defined for all pillars with explicit boolean array')
                else:
                    print('geometry is not defined for all pillars')
                    cells_all_defined = False
    cell_is_def_root = grid.resolve_geometry_child('CellGeometryIsDefined')
    cells_all_defined = False
    if cell_is_def_root is None:
        print('grid does not include boolean array indicating which cell geometries are defined!')
    else:
        cell_is_def_type = rqet.node_type(cell_is_def_root)
        if cell_is_def_type == 'BooleanConstantArray':
            cells_all_defined = grid.cell_geometry_is_defined()
            assert cells_all_defined is not None
            if cells_all_defined:
                print('geometry is defined for all cells')
            else:
                print('geometry is not defined for any cells!')
        else:
            assert cell_is_def_type == 'BooleanHdf5Array'
            if detail_level > 1:
                cell_is_def_array = grid.cell_geometry_is_defined_ref()
                if grid.geometry_defined_for_all_cells_cached is None or not grid.geometry_defined_for_all_cells_cached:
                    assert cell_is_def_array is not None
                    cells_defined = np.count_nonzero(cell_is_def_array)
                    print('geometry is defined for {0:1d} cells ({1:4.2f}%)'.format(
                        cells_defined, 100.0 * float(cells_defined) / float(cell_count)))
                    cells_all_defined = (cells_defined == cell_count)
                else:
                    cells_all_defined = True
                if cells_all_defined:
                    print('geometry is defined for all cells with explicit boolean array')
    points_root = grid.resolve_geometry_child('Points')
    if points_root is None:
        print('grid does not include (corner) points data!')


#   elif cells_all_defined:
    else:
        laziness = (detail_level < 2)
        for local in [True, False]:
            xyz_box = grid.xyz_box(points_root = points_root, lazy = laziness, local = local)
            if local:
                local_str = 'local'
            else:
                local_str = 'global'
            if laziness:
                laziness_str = 'lazy'
            else:
                laziness_str = 'thorough'
            print(local_str + ' bounding values of geometry min, max (' + laziness_str + '):')
            for xyz in range(3):
                print(' {0}: {1:12.3f} {2:12.3f}'.format('xyz'[xyz], xyz_box[0, xyz], xyz_box[1, xyz]))
    # todo: check whether all HDF5 refs are to same file

hdf5_dataset_count = 0


def print_hdf5_line(name, group_or_dataset):
    """Prints information about one dataset (array) in an hdf5 file.

    :meta private:
    """

    global hdf5_dataset_count
    if isinstance(group_or_dataset, h5py.Group):
        return None
    print(name + '  ' + str(group_or_dataset.shape) + '  ' + str(group_or_dataset.dtype))
    hdf5_dataset_count += 1
    return None


def print_EpcExternalPartReference(model, node, detail_level = 0):
    """Prints information about an hdf5 external part."""

    global hdf5_dataset_count
    print('mime type: ' + str(rqet.node_text(rqet.find_tag(node, 'MimeType'))))
    if not detail_level or 'uuid' not in node.attrib.keys():
        return
    uuid = bu.uuid_from_string(node.attrib['uuid'])
    file_name = model.h5_file_name(uuid)
    print('hdf5 file: ' + str(file_name))
    hdf5 = model.h5_access(uuid)
    hdf5_dataset_count = 0
    if hdf5 is not None:
        for (key, group) in hdf5.items():
            print('*** ' + str(key) + ': ' + str(group))
            group.visititems(print_hdf5_line)
    print('number of data sets (arrays): ' + str(hdf5_dataset_count))


def print_TimeSeries(model, node, detail_level = 0):
    """Prints information form a time series xml tree."""

    if node is None:
        return
    print('number of timestamps: ' + str(rqet.count_tag(node, 'Time')))
    if not detail_level:
        return
    time_series = rts.any_time_series(model, uuid = node.attrib['uuid'])
    for index in range(time_series.number_of_timestamps()):
        if time_series.timeframe == 'geologic':
            print(f'{index:>5d}  {time_series.timestamp(index)}')
        elif index:
            print('{0:>5d}  {1} {2:>5d}'.format(index, rts.simplified_timestamp(time_series.timestamp(index)),
                                                time_series.step_days(index)))
        else:
            print('{0:>5d}  {1}  step (days)'.format(0, rts.simplified_timestamp(time_series.timestamp(0))))


def print_StringTableLookup(model, node, detail_level = 0):
    """Prints information from a string lookup table xml tree."""

    entries_node_list = rqet.list_of_tag(node, 'Value')
    if entries_node_list is None or len(entries_node_list) == 0:
        print('no entries in lookup table')
        return
    print(str(len(entries_node_list)) + ' entries in lookup table:')
    if not detail_level:
        return
    entries_pair_list = []
    for entry_node in entries_node_list:
        key_node = rqet.find_tag(entry_node, 'Key')
        if key_node is None:
            continue
        key_type = rqet.node_type(key_node)
        if key_type == 'integer':
            key = int(key_node.text)
        else:
            key = key_node.text
        value = rqet.node_text(rqet.find_tag(entry_node, 'Value'))
        entries_pair_list.append((key, value))
    entries_pair_list.sort()
    for (key, value) in entries_pair_list:
        if isinstance(key, int):
            print('{0:>4d}: {1}'.format(key, value))
        else:
            print(str(key) + ': ' + str(value))


def print_reference_node_and_return_referenced_part(node, heading):
    """Prints some information about a part being referenced in an xml node."""

    if node is None:
        print(str(heading) + ': (missing)')
        return None
    print(str(heading) + ':')
    print('   title: ' + str(rqet.node_text(rqet.find_tag(node, 'Title'))))
    content_type = rqet.content_type(rqet.node_text(rqet.find_tag(node, 'ContentType')))
    if content_type is not None:
        print('   type:  ' + str(rcd.readable_class(content_type)))
    uuid_str = rqet.node_text(rqet.find_tag(node, 'UUID'))
    if uuid_str is not None:
        print('   uuid:  ' + str(uuid_str))
    if content_type is None or content_type[:4] != 'obj_' or uuid_str is None:
        return None
    if uuid_str[0] == '_':
        uuid_str = uuid_str[1:]  # tolerate fesapi quirk
    return content_type + '_' + uuid_str + '.xml'


def print_Property(model, node, detail_level = 0):
    """Prints information about a grid property, mostly from the xml tree."""

    property_kind = None
    property_kind_node = rqet.find_tag(node, 'PropertyKind')
    if property_kind_node is None:
        print('property kind is missing!')
    else:
        kind_node = rqet.find_tag(property_kind_node, 'Kind')
        if kind_node is not None:
            property_kind = rqet.node_text(kind_node)
            print('kind: ' + str(property_kind))
        else:
            local_kind_node = rqet.find_tag(property_kind_node, 'LocalPropertyKind')
            if local_kind_node is not None:
                print('kind: ' + str(rqet.find_tag_text(local_kind_node, 'Title')))
            else:
                print('property kind xml not recognised!')
    facet_type = None
    facet = None
    facet_node_list = rqet.list_of_tag(node, 'Facet')
    if facet_node_list is not None:
        for facet_node in facet_node_list:
            print('facet: ' + str(rqet.node_text(rqet.find_tag(facet_node, 'Facet'))) + ': ' +
                  str(rqet.find_tag_text(facet_node, 'Value')))
            if not facet_type:  # arbitrarily pick up first facet in list to use in helping to guess uom, if needed
                facet_type = rqet.node_text(rqet.find_tag(facet_node, 'Facet'))
                facet = rqet.node_text(rqet.find_tag(facet_node, 'Value'))
    citation_title = rqet.node_text(rqet.find_tag(rqet.find_tag(node, 'Citation'), 'Title'))
    if citation_title is not None:
        (derived_property_kind, derived_facet_type,
         derived_facet) = rprop.property_kind_and_facet_from_keyword(citation_title)
        if derived_property_kind is not None:
            if property_kind is None:
                property_kind = derived_property_kind
            if derived_facet_type is None:
                print('(derived kind: ' + derived_property_kind + ')')
            else:
                print('(derived kind: ' + str(derived_property_kind) + '; facet: ' + str(derived_facet_type) + ': ' +
                      str(derived_facet) + ')')
    print('realization: ' + optional(rqet.find_tag_int(node, 'RealizationIndex')))
    grid = None
    time_series = time_series_uuid = time_series_part = None
    time_index_node = rqet.find_tag(node, 'TimeIndex')
    if time_index_node is None:
        print('static property (no time index)')
    else:
        index_node = rqet.find_tag(time_index_node, 'Index')
        print('time index: ' + str(rqet.node_text(index_node)))
        if index_node is None:
            time_index = None
        else:
            time_index = int(index_node.text)
        time_series_ref_node = rqet.find_tag(time_index_node, 'TimeSeries')
        time_series_part = print_reference_node_and_return_referenced_part(time_series_ref_node, 'time series')
        if detail_level > 0:
            time_series_uuid = model.uuid_for_part(time_series_part)
            if time_series_uuid is None:
                print('   time series part not found!')
                log.warning('missing time series part: ' + str(time_series_part))
            else:
                time_series = rts.any_time_series(model, uuid = time_series_uuid)
                print('   number of timestamps in series: ' + str(time_series.number_of_timestamps()))
            if time_index is not None and time_series is not None:
                print('timestamp for this part: ' + str(rts.simplified_timestamp(time_series.timestamp(time_index))))
    minimum = rqet.node_text(rqet.find_tag(node, 'MinimumValue'))
    print('minimum (xml): ' + str(minimum))
    maximum = rqet.node_text(rqet.find_tag(node, 'MaximumValue'))
    print('maximum (xml): ' + str(maximum))
    units = rqet.find_tag_text(node, 'UOM')
    if units is not None:
        print('units: ' + units)
        if units in ['unknown', 'Euc'] and property_kind is not None and detail_level > 0:
            guessed_uom = rprop.guess_uom(property_kind, minimum, maximum, grid, facet_type = facet_type, facet = facet)
            if guessed_uom:
                print('(guessed units: ' + str(guessed_uom) + ')')
    print('count: ' + str(rqet.find_tag_text(node, 'Count')))
    print('indexable element: ' + str(rqet.find_tag_text(node, 'IndexableElement')))
    values_node = rqet.find_tag(node, 'PatchOfValues')
    if values_node is None:
        print('patch of values is missing!')
        return
    print('patch of values is present in xml')
    if detail_level < 2 or grid is None:
        return
    single_prop_collection = rprop.GridPropertyCollection()
    single_prop_collection.set_grid(grid)
    part_name = rqet.part_name_for_part_root(node)
    single_prop_collection.add_part_to_dict(part_name,
                                            trust_uom = units is not None and units not in ['', 'unknown', 'Euc'])
    cached_array = single_prop_collection.cached_part_array_ref(part_name)
    if cached_array is None:
        print('failed to cache array for part ' + str(part_name))
        return None
    print('type of array elements: ' + str(cached_array.dtype))
    print('shape of array (nk, nj, ni): ' + str(cached_array.shape))
    print('number of elements in array: ' + str(cached_array.size))
    non_zero_count = np.count_nonzero(cached_array)
    print('number of non-zero elements: ' + str(non_zero_count))
    print('minimum (data, all cells): ' + str(np.nanmin(cached_array)))
    print('maximum (data, all cells): ' + str(np.nanmax(cached_array)))
    if cached_array.dtype == 'float':
        average = np.mean(cached_array)  # todo: handle NaN and Inf in a better way
        print('mean value (all cells): ' + str(average))
        if property_kind in ['rock volume', 'pore volume']:
            print('sum of values (all cells): ' + str(np.sum(cached_array)))
    if grid and grid.inactive is not None:
        masked_array = single_prop_collection.cached_part_array_ref(part_name, masked = True)
        inactive_count = np.count_nonzero(masked_array.mask)
        print('active cell count: ' + str(cached_array.size - inactive_count))
        print('inactive cell count: ' + str(inactive_count))
        print('minimum (data, active cells): ' + str(np.nanmin(masked_array)))
        print('maximum (data, active cells): ' + str(np.nanmax(masked_array)))
        if cached_array.dtype == 'float':
            average = np.mean(masked_array)  # todo: handle NaN and Inf in a better way
            print('mean value (active cells): ' + str(average))
            if property_kind in ['rock volume', 'pore volume']:
                print('sum of values (active cells): ' + str(np.nansum(masked_array)))
    single_prop_collection.uncache_part_array(part_name)


def print_CategoricalProperty(model, node, detail_level = 0):
    """Prints information about a categorical grid property, mostly from the xml tree."""

    print_Property(model, node, detail_level = detail_level)
    lookup_node = rqet.find_tag(node, 'Lookup')
    if lookup_node is None:
        print('no lookup table referenced')
    else:
        print_reference_node_and_return_referenced_part(lookup_node, 'lookup table reference')


def print_ContinuousProperty(model, node, detail_level = 0):
    """Prints information about a continuous (float) grid property, mostly from the xml tree."""

    print_Property(model, node, detail_level = detail_level)


def print_DiscreteProperty(model, node, detail_level = 0):
    """Prints information about a discrete (integer) grid property, mostly from the xml tree."""

    print_Property(model, node, detail_level = detail_level)


def print_TriangulatedSetRepresentation(model, node, detail_level = 0):
    """Prints information about a triangulated set representation (surface) from the xml tree."""

    print('surface role: ' + required(rqet.find_tag_text(node, 'SurfaceRole')))
    patch_count = rqet.count_tag(node, 'TrianglePatch')
    print('patch count:  ' + str(patch_count))
    if detail_level > 0:
        for patch in rqet.list_of_tag(node, 'TrianglePatch'):
            print('   patch index:     ' + required(rqet.find_tag_int(patch, 'PatchIndex')))
            print('   triangle count:  ' + required(rqet.find_tag_int(patch, 'Count')))
            print('   node count:      ' + required(rqet.find_tag_int(patch, 'NodeCount')))
            if detail_level > 1:
                print('   node null value: ' + optional(rqet.find_tag_int(patch, 'NullValue')))


def print_PointSetRepresentation(model, node, detail_level = 0):
    """Prints information about a point set representation from the xml tree."""

    patch_count = rqet.count_tag(node, 'NodePatch')
    print('patch count:  ' + str(patch_count))
    if detail_level > 0:
        total_count = 0
        for patch in rqet.list_of_tag(node, 'NodePatch'):
            print('   patch index: ' + required(rqet.find_tag_int(patch, 'PatchIndex')))
            point_count = rqet.find_tag_int(patch, 'Count')
            print('   point count: ' + required(point_count))
            if point_count:
                total_count += point_count
        print('(total point count: {})'.format(total_count))


def print_GridConnectionSetRepresentation(model, node, detail_level = 0):
    """Prints information about a grid connection set representation (fault cell face set), mostly from the xml tree."""

    print('face count: ' + required(rqet.find_tag_int(node, 'Count')))
    ci_node = rqet.find_tag(node, 'ConnectionInterpretations')
    if ci_node is None:
        print('(connection interpretations missing)')
        return
    feature_count = rqet.count_tag(ci_node, 'FeatureInterpretation')
    print('feature (fault or fracture) count: ' + str(feature_count))
    if detail_level > 0:
        if detail_level > 1:
            gcs = rqf.GridConnectionSet(model, connection_set_root = node)
        feature_index = 0
        for feature in rqet.list_of_tag(ci_node, 'FeatureInterpretation'):
            print_reference_node_and_return_referenced_part(feature, 'feature interpretation reference')
            if detail_level > 1:
                face_count_per_feature = len(gcs.raw_list_of_cell_face_pairs_for_feature_index(feature_index)[0])
                print('   face pair count: ' + str(face_count_per_feature))
            feature_index += 1


def print_MdDatum(model, node, detail_level = 0):
    """Prints information about a measured depth datum, from the xml tree."""

    print('reference: ' + required(rqet.find_tag_text(node, 'MdReference')))
    location_xyz = point_3d(node, 'Location')
    print('location:  ' + format_xyz(location_xyz))
    if detail_level > 0:
        crs_ref = rqet.find_tag(node, 'LocalCrs')
        crs_part = print_reference_node_and_return_referenced_part(crs_ref,
                                                                   'local coordinate reference system reference')
        if detail_level > 1:
            crs = rqc.Crs(model, uuid = rqet.uuid_in_part_name(crs_part))
            global_xyz = crs.local_to_global(location_xyz)
            print('global location: ' + format_xyz(global_xyz))


def print_WellboreFrameRepresentation(model, node, detail_level = 0):
    """Prints information about a wellbore frame representation, from the xml tree."""

    print('md mode count: ' + required(rqet.find_tag_int(node, 'NodeCount')))
    trajectory_ref = rqet.find_tag(node, 'Trajectory')
    print_reference_node_and_return_referenced_part(trajectory_ref, 'wellbore trajectory reference')


def print_WellboreMarkerFrameRepresentation(model, node, detail_level = 0):
    """Prints information about a wellbore marker frame representation, from the xml tree."""

    print_WellboreFrameRepresentation(model, node, detail_level = detail_level)
    # todo: add details
    marker_count = len(rqet.list_of_tag(node, 'WellboreMarker'))
    print('marker count: ' + str(marker_count))
    print('(other details not yet available for this type of object)')


def print_BlockedWellboreRepresentation(model, node, detail_level = 0):
    """Prints information about a blocked wellbore representation, from the xml tree."""

    print_WellboreFrameRepresentation(model, node, detail_level = detail_level)
    print('cell count: ' + required(rqet.find_tag_int(node, 'CellCount')))
    print('grid count: ' + required(rqet.count_tag(node, 'Grid')))
    if detail_level > 0:
        for grid_ref_node in rqet.list_of_tag(node, 'Grid'):
            print_reference_node_and_return_referenced_part(grid_ref_node, 'grid reference')


def print_WellboreTrajectoryRepresentation(model, node, detail_level = 0):
    """Prints information about a wellbore trajectory representation, from the xml tree."""

    lk_dict = {
        -1: 'null - no line',
        0: 'vertical)',
        1: 'linear spline',
        2: 'natural cubic spline',
        3: 'cubic spline',
        4: 'z linear cubic spline',
        5: 'minimum-curvature spline'
    }
    unit_str = rqet.find_tag_text(node, 'MdUom')
    if unit_str is None:
        unit_str = '(missing units)'
    print('start md:  ' + required(rqet.find_tag_float(node, 'StartMd')) + ' ' + unit_str)
    print('finish md: ' + required(rqet.find_tag_float(node, 'FinishMd')) + ' ' + unit_str)
    print('md units:  ' + unit_str)
    md_ref_node = rqet.find_tag(node, 'MdDatum')
    if md_ref_node is None:
        print('(missing measured depth datum reference)')
    else:
        print_reference_node_and_return_referenced_part(md_ref_node, 'measured depth datum reference')
    geom_node = rqet.find_tag(node, 'Geometry')
    if geom_node is None:
        print('(presumed vertical - no geometry)')
        return
    line_kind = rqet.find_tag_int(geom_node, 'LineKindIndex')
    if line_kind is None:
        lk_str = '(missing)'
    elif line_kind >= -1 and line_kind <= 5:
        lk_str = str(line_kind) + ' (' + lk_dict[line_kind] + ')'
    else:
        lk_str = str(line_kind) + ' (invalid value)'
    print('line kind:  ' + lk_str)
    print('knot count: ' + required(rqet.find_tag_int(geom_node, 'KnotCount')))
    tv_node = rqet.find_tag(geom_node, 'TangentVectors')
    if tv_node:
        print('(tangent vectors are present)')
    else:
        print('(optional tangent vectors are not present)')
    print('domain:     ' + optional(rqet.find_tag_text(node, 'MdDomain')))
    ds_ref_node = rqet.find_tag(node, 'DeviationSurvey')
    if detail_level > 0:
        crs_ref = rqet.find_tag(geom_node, 'LocalCrs')
        print_reference_node_and_return_referenced_part(crs_ref, 'local coordinate reference system reference')
    if ds_ref_node is not None:
        print_reference_node_and_return_referenced_part(ds_ref_node, 'deviation survey reference')
    if detail_level > 1:
        try:
            trajectory = rqw.Trajectory(model, trajectory_root = node)
            print('min xyz: ' + str(np.amin(trajectory.control_points, axis = 0)))
            print('max xyz: ' + str(np.amax(trajectory.control_points, axis = 0)))
            print('    x        y       z')
            for (x, y, z) in trajectory.control_points:
                print(f' {x:8.1f} {y:8.1f} {z:7.1f}')
        except Exception:
            print('(problem displaying control points)')
            log.exception('trajectory issue')


def print_TectonicBoundaryFeature(model, node, detail_level = 0):
    """Prints information about a tectonic boundary feature (fault or fracture), from the xml tree."""

    print('tectonic boundary kind: ' + required(rqet.find_tag_text(node, 'TectonicBoundaryKind')))


def print_GeneticBoundaryFeature(model, node, detail_level = 0):
    """Prints information about a genetic boundary feature (horizon or geobody boundary), from the xml tree."""

    print('genetic boundary kind: ' + required(rqet.find_tag_text(node, 'GeneticBoundaryKind')))
    # todo: optional absolute age


def print_WellboreFeature(model, node, detail_level = 0):
    """Prints information about a wellbore feature, from the xml tree."""

    pass  # nothing but citation block in this feature class


def print_OrganizationFeature(model, node, detail_level = 0):
    """Prints information about an organization feature, from the xml tree."""

    pass  # nothing but citation block in this feature class


def print_FaultInterpretation(model, node, detail_level = 0):
    """Prints information about a fault interpretation, from the xml tree."""

    print('domain:        ' + required(rqet.find_tag_text(node, 'Domain')))
    print('is listric:    ' + optional(rqet.find_tag_text(node, 'IsListric')))
    print('maximum throw: ' + optional(rqet.find_tag_float(node, 'MaximumThrow')))
    print('mean azimuth:  ' + optional(rqet.find_tag_float(node, 'MeanAzimuth')))
    print('mean dip:      ' + optional(rqet.find_tag_float(node, 'MeanDip')))
    throw_understanding = rqet.find_tag_text(node, 'ThrowInterpretation')
    if not throw_understanding:
        throw_understanding = '(presumed normal)'
    print('throw interpretation: ' + throw_understanding)


def print_HorizonInterpretation(model, node, detail_level = 0):
    """Prints information about a horizon interpretation, from the xml tree."""

    print('domain: ' + required(rqet.find_tag_text(node, 'Domain')))
    br_node_list = rqet.list_of_tag(node, 'BoundaryRelation')
    if len(br_node_list):
        for br_node in br_node_list:
            print('boundary relation: ' + required(br_node.text))
    else:
        print('(no boundary relation(s) specified)')
    sss_node_list = rqet.list_of_tag(node, 'SequenceStratigraphySurface')
    if len(sss_node_list):
        for sss_node in sss_node_list:
            print('sequence stratigraphy surface: ' + required(sss_node.text))
    else:
        print('(no sequence stratigraphy surface(s) specified)')


def print_WellboreInterpretation(model, node, detail_level = 0):
    """Prints information about a wellbore interpretation, from the xml tree."""

    print('is drilled: ' + required(rqet.find_tag_bool(node, 'IsDrilled')))


def print_EarthModelInterpretation(model, node, detail_level = 0):
    """Prints information about an earth model interpretation, from the xml tree."""

    print('domain: ' + required(rqet.find_tag_text(node, 'Domain')))


def print_node(model, node, detail_level = 0):
    """Prints information about an object from an xml node."""

    if node is None:
        return
    title = rqet.citation_title_for_node(node)
    if title:
        print('title: ' + str(title))
    obj_type = rqet.node_type(node)
    if obj_type is None:
        return
    print('object type: ' + str(rcd.readable_class(obj_type)))
    if 'uuid' in node.attrib.keys():
        print('uuid: ' + str(node.attrib['uuid']))
    if obj_type.endswith('Property'):
        support_node = rqet.find_tag(node, 'SupportingRepresentation')
        if support_node is None:
            print('no supporting representation referenced')
        else:
            print_reference_node_and_return_referenced_part(support_node, 'supporting representation reference')
    elif obj_type.endswith('Representation'):
        interp_node = rqet.find_tag(node, 'RepresentedInterpretation')
        if interp_node is None:
            print('no represented interpretation referenced')
        else:
            print_reference_node_and_return_referenced_part(interp_node, 'represented interpretation reference')
    elif obj_type.endswith('Interpretation'):
        feature_node = rqet.find_tag(node, 'InterpretedFeature')
        if feature_node is None:
            print('no interpreted feature referenced')
        else:
            print_reference_node_and_return_referenced_part(feature_node, 'interpreted feature reference')
    if obj_type == 'obj_LocalDepth3dCrs':
        print_LocalDepth3dCrs(model, node)
    elif obj_type == 'obj_LocalTime3dCrs':
        print_LocalTime3dCrs(model, node)
    elif obj_type == 'obj_IjkGridRepresentation':
        print_IjkGridRepresentation(model, node, detail_level = detail_level)
    elif obj_type == 'obj_EpcExternalPartReference':
        print_EpcExternalPartReference(model, node, detail_level = detail_level)
    elif obj_type == 'obj_TimeSeries':
        print_TimeSeries(model, node, detail_level = detail_level)
    elif obj_type == 'obj_StringTableLookup':
        print_StringTableLookup(model, node, detail_level = detail_level)
    elif obj_type == 'obj_ContinuousProperty':
        print_ContinuousProperty(model, node, detail_level = detail_level)
    elif obj_type == 'obj_DiscreteProperty':
        print_DiscreteProperty(model, node, detail_level = detail_level)
    elif obj_type == 'obj_CategoricalProperty':
        print_CategoricalProperty(model, node, detail_level = detail_level)
    elif obj_type == 'obj_TriangulatedSetRepresentation':
        print_TriangulatedSetRepresentation(model, node, detail_level = detail_level)
    elif obj_type == 'obj_PointSetRepresentation':
        print_PointSetRepresentation(model, node, detail_level = detail_level)
    elif obj_type == 'obj_GridConnectionSetRepresentation':
        print_GridConnectionSetRepresentation(model, node, detail_level = detail_level)
    elif obj_type == 'obj_MdDatum':
        print_MdDatum(model, node, detail_level = detail_level)
    elif obj_type == 'obj_WellboreFrameRepresentation':
        print_WellboreFrameRepresentation(model, node, detail_level = detail_level)
    elif obj_type == 'obj_WellboreMarkerFrameRepresentation':
        print_WellboreMarkerFrameRepresentation(model, node, detail_level = detail_level)
    elif obj_type == 'obj_BlockedWellboreRepresentation':
        print_BlockedWellboreRepresentation(model, node, detail_level = detail_level)
    elif obj_type == 'obj_WellboreTrajectoryRepresentation':
        print_WellboreTrajectoryRepresentation(model, node, detail_level = detail_level)
    elif obj_type == 'obj_TectonicBoundaryFeature':
        print_TectonicBoundaryFeature(model, node, detail_level = detail_level)
    elif obj_type == 'obj_GeneticBoundaryFeature':
        print_GeneticBoundaryFeature(model, node, detail_level = detail_level)
    elif obj_type == 'obj_WellboreFeature':
        print_WellboreFeature(model, node, detail_level = detail_level)
    elif obj_type == 'obj_FaultInterpretation':
        print_FaultInterpretation(model, node, detail_level = detail_level)
    elif obj_type == 'obj_HorizonInterpretation':
        print_HorizonInterpretation(model, node, detail_level = detail_level)
    elif obj_type == 'obj_WellboreInterpretation':
        print_WellboreInterpretation(model, node, detail_level = detail_level)
    elif obj_type == 'obj_OrganizationFeature':
        print_OrganizationFeature(model, node, detail_level = detail_level)
    elif obj_type == 'obj_EarthModelInterpretation':
        print_EarthModelInterpretation(model, node, detail_level = detail_level)
        # todo: other object types
    else:
        print('(details unavailable for this type of object)')
    print_citation(node)



"""schema.py: Functions for working with Energistics schema definitions.

To run as a script, set the following 3 environment variables and a file will be created with rst-format tables
listing the elements for each obj_* class:

    EML_SCHEMA_SPECIFIC_DIR: directory containing main xsd files for the schema
    EML_SCHEMA_COMMON_DIR: directory containing common xsd files
    EML_RST_TABLE_FILE: output filename (will be overwritten) for the tables
"""

# note: some parts of this module require the main class names to begin 'obj_'
# this will need furher work for some standards other than RESQML 2.0.1

is_main = (__name__ == '__main__')

import os
# from lxml.etree import parse
from xml.etree.ElementTree import parse
import glob

# the xsd files are xml files, so we reuse resqpy xml processing functions here to look at the schema
import resqpy.olio.xml_et as rqet


def xsd_to_dicts(schema_dirs):
    """Reads schema definition xsd files and returns dictionaries.

    arguments:
        schema_dirs (list of str): directories containing the schema definition files

    returns:
        (simple_types, complex_types) each being a dictionary;
        simple types: dict of name -> (restriction_base, list of valid (value, value_doc), type_doc);
        complex types: dict of name -> (extension_base, list of elements, type_doc);
        where element is (tag_name, type, min_occurs, max_occurs)

    notes:
        restriction_base, extension_base, tag_name and type items are strings;
        type_doc and value_doc items are the human readable strings from the xsd annotation documentation fields
    """

    def _process_simple_types(schema, simple_types):
        for simpleton in rqet.list_of_tag(schema, 'simpleType'):
            name = simpleton.attrib['name']
            doc = rqet.find_nested_tags_text(simpleton, ['annotation', 'documentation'])
            restriction = rqet.find_tag(simpleton, 'restriction')
            restrict = None if restriction is None else rqet.stripped_of_prefix(restriction.attrib['base'])
            value_list = []
            enum_list = rqet.list_of_tag(restriction, 'enumeration')
            if enum_list is not None:
                for enumeration in enum_list:
                    value = enumeration.attrib['value']
                    v_doc = rqet.find_nested_tags_text(enumeration, ['annotation', 'documentation'])
                    if v_doc is not None:
                        v_doc = str(v_doc).replace('\n', ' ')
                    value_list.append((value, v_doc))
            simple_types[name] = (restrict, value_list, doc)

    def _process_complex_types(schema, complex_types):
        for completon in rqet.list_of_tag(schema, 'complexType'):
            name = completon.attrib['name']
            doc = rqet.find_nested_tags_text(completon, ['annotation', 'documentation'])
            content = rqet.find_tag(completon, 'complexContent')
            extension = rqet.find_tag(content, 'extension')
            if extension is None:
                e_base = None
            else:
                e_base = extension.attrib['base']
                content = extension
            if content is None:
                content = completon
            sequence = rqet.find_tag(content, 'sequence')
            if sequence is None:
                sequence = content
            element_list = []
            e_nodes = rqet.list_of_tag(sequence, 'element')
            if e_nodes is not None:
                for element in e_nodes:
                    e_name = element.attrib['name']
                    flavour = element.attrib['type']  # could strip prefix here?
                    min_occurs = element.get('minOccurs')
                    max_occurs = element.get('maxOccurs')
                    element_list.append((e_name, flavour, min_occurs, max_occurs))
            complex_types[name] = (e_base, element_list, doc)

    simple_types = {}
    complex_types = {}

    for directory in schema_dirs:
        xsd_list = glob.glob(os.path.join(directory, '*.xsd'))
        for xsd in xsd_list:
            xml_tree = parse(xsd)
            schema = xml_tree.getroot()
            _process_simple_types(schema, simple_types)
            _process_complex_types(schema, complex_types)

    return (simple_types, complex_types)


def expand_complex_types(complex_types):
    """Expands the elements within the complex types and returns a flattened dictionary.

    arguments:
        complex_types (dict): as returned by xsd_to_dicts()[1]

    returns:
        flattened_types: dict of name -> list of elements;
        where element is (tag_name, type, min_occurs, max_occurs)

    notes:
        the elements returned will be simple types or abstract types; an abstract type indicates
        a choice of possible data sub-structures
    """

    def _expand(key, complex_types, expansion_state, flattened_types):
        # recursively expand extension base elements
        if expansion_state[key] > 0:
            return
        expansion_state[key] = 1
        e_base, element_list, _ = complex_types[key]
        if e_base is not None:
            e_base = e_base.split(sep = ':')[-1]
            _expand(e_base, complex_types, expansion_state, flattened_types)
            assert expansion_state[e_base] != 1, f'circular reference in {e_base}'
            element_list = flattened_types[e_base] + element_list
        flattened_types[key] = element_list
        expansion_state[key] = 2

    flattened_types = {}
    expansion_state = {}  # tag -> int; 0: not expanded; 1: in progress; 2: expanded
    for key in complex_types.keys():
        expansion_state[key] = 0

    for name in complex_types.keys():
        _expand(name, complex_types, expansion_state, flattened_types)

    return flattened_types


def obj_names(complex_types):
    """Returns list of 'obj_*' high level object type names, sorted alphabetically.

    arguments:
        complex_types (dict): as returned by xsd_to_dicts()[1]

    returns:
        list of str
    """

    return sorted([k for k in complex_types.keys() if k.startswith('obj_')])


def print_flattened(complex_types, flattened_types, type_key, max_levels = 99, indent = 0):
    """Recursively prints a list of the (flattened) elements for a complex type.

    arguments:
        flattened_types (dict): as returned by expand_complex_types()
        type_key (str): the name of the complex type for which an element list print is required
        max_levels (int): the maximum number of nested levels to show in printout
        indent (int, default 0): initial indentation level
    """

    if indent > max_levels:
        return
    for e in flattened_types[type_key]:
        print(' ' * 3 * indent, f'{e[0]}: ({e[1]}) ', end = '')
        if e[2] == '0' and e[3] == '1':
            print('optional')
        elif e[3] == 'unbounded':
            print(f'{e[2]} or more')
        elif e[2] == '1' and e[3] == '1':
            print()
        else:
            print(f'{e[2]} to {e[3]}')
        if e[1].startswith('resqml:'):
            ref_key = e[1].split(':')[-1]
            if ref_key in complex_types.keys():
                print_flattened(complex_types, flattened_types, ref_key, max_levels = max_levels, indent = indent + 1)


def manifestations(complex_types, abstract_type, is_obj = False):
    """Returns a list of type names which are manifestations of a given abstract type.

    arguments:
        complex_types (dict): as returned by xsd_to_dicts()[1]
        abstract_type (str): the name of the abstract type of interest
        is_obj (bool, default False): if True, returns a list of obj_* types that are direct
            extensions of the abstract type; if False, returns a list of the (non obj_*) types
            which are the alternative representations of the abstract type

    returns:
        list of str
    """

    m_list = []

    for t in complex_types.keys():
        e_base = complex_types[t][0]
        if e_base is not None and e_base.split(sep = ':')[-1] == abstract_type and t.startswith('obj_') == is_obj:
            m_list.append(t)

    return m_list


def is_abstract_type(type_name):
    """Returns True if type_name appears to be the name of an abstract type; False otherwise."""

    return type_name.startswith('Abstract')


def abstract_types(complex_types):
    """Returns list of abstract types found amongst complex types."""

    return [k for k in complex_types.keys() if is_abstract_type(k)]


def max_sizes_for_rst(flattened_types, complex_types, key = None):
    """Returns a list of 3 integers being the required column widths for rst format tables."""

    def _max_sizes(flattened_types, complex_types, key, indent = 0, max_so_far = [0, 0, 0]):
        for e in flattened_types[key]:
            s0 = 3 * indent + len(e[0])
            if s0 > max_so_far[0]:
                max_so_far[0] = s0
            s1 = len(e[1])
            if s1 > max_so_far[1]:
                max_so_far[1] = s1
            if e[1].startswith('resqml:'):
                ref_key = e[1].split(':')[-1]
                if ref_key in complex_types:
                    _max_sizes(flattened_types, complex_types, ref_key, indent = indent + 1, max_so_far = max_so_far)
        return max_so_far

    ms = [0, 0, 10]

    key_list = obj_names(complex_types) if key is None else [key]
    for key in key_list:
        ms = _max_sizes(flattened_types, complex_types, key, max_so_far = ms)

    return ms


def write_rst_table(fp, flattened_types, complex_types, key, max_sizes, header = True, max_levels = 99, sort = False):
    """Writes an rst format table of elements for a complex type, to an already opened file."""

    def _write_bar(fp, max_sizes):
        line = '+' + '-' * max_sizes[0] + '+' + '-' * max_sizes[1] + '+' + '-' * max_sizes[2] + '+\n'
        fp.write(line)

    def _pf_table(fp, flattened_types, complex_types, key, ms, max_levels, indent = 0, sort = False):
        if indent > max_levels:
            return
        table_width = ms[0] + ms[1] + ms[2] + 4
        elements = flattened_types[key]
        if sort:
            elements = sorted(elements)
        for e in elements:
            line = '|' + '·' * 3 * indent + e[0]
            if len(line) < ms[0] + 1:
                line += ' ' * (ms[0] + 1 - len(line))
            line += '|' + e[1]
            if len(line) < ms[0] + ms[1] + 2:
                line += ' ' * (ms[0] + ms[1] + 2 - len(line))
            line += '|'
            if e[2] == '0' and e[3] == '1':
                line += 'optional'
            elif e[3] == 'unbounded':
                line += f'{e[2]} or more'
            elif e[2] == '1' and e[3] == '1':
                line += 'required'
            else:
                line += f'{e[2]} to {e[3]}'
            if len(line) < table_width - 1:
                line += ' ' * (table_width - 1 - len(line))
            line += '|\n'
            fp.write(line)
            _write_bar(fp, ms)
            if e[1].startswith('resqml:'):
                ref_key = e[1].split(':')[-1]
                if ref_key in complex_types.keys():
                    _pf_table(fp,
                              flattened_types,
                              complex_types,
                              ref_key,
                              ms,
                              max_levels = max_levels,
                              indent = indent + 1,
                              sort = sort)

    if header:
        fp.write(f'\n{key}\n')
        fp.write('-' * len(key) + '\n\n')

    _write_bar(fp, max_sizes)
    _pf_table(fp, flattened_types, complex_types, key, max_sizes, max_levels, sort = sort)


def write_all_obj_rst_tables(flattened_types, complex_types, file_name, mode = 'a', sort_elements = False):
    """Writes a sequence of tables in rst format, one for each obj_* class, to a named file."""

    max_sizes = max_sizes_for_rst(flattened_types, complex_types)

    with open(file_name, mode) as fp:
        for key in obj_names(complex_types):
            write_rst_table(fp, flattened_types, complex_types, key, max_sizes, sort = sort_elements)


# if running as a script, check and use 3 environment variable settings
if is_main:
    schema_main_dir = os.environ.get('EML_SCHEMA_SPECIFIC_DIR')
    schema_common_dir = os.environ.get('EML_SCHEMA_COMMON_DIR')
    table_file = os.environ.get('EML_RST_TABLE_FILE')
    if schema_main_dir is None or schema_common_dir is None or table_file is None:
        print('set the following 3 environment variables before running this script:')
        for env_var in ['EML_SCHEMA_SPECIFIC_DIR', 'EML_SCHEMA_COMMON_DIR', 'EML_RST_TABLE_FILE']:
            print(f'   {env_var}')
        raise ValueError('environment variable not set')
    simple_types, complex_types = xsd_to_dicts([schema_main_dir, schema_common_dir])
    flattened_types = expand_complex_types(complex_types)
    write_all_obj_rst_tables(flattened_types, complex_types, table_file, mode = 'w', sort_elements = False)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# See also https://wwoods.github.io/2016/06/09/easy-sphinx-documentation-without-the-boilerplate/#setup

# Nb. useful stackexchange thread about the :recursive: option and custom templates:
# https://stackoverflow.com/questions/2701998/sphinx-autodoc-is-not-automatic-enough/62613202#62613202

# -- Path setup --------------------------------------------------------------

# Do not do sys.path surgery
# Instead, resqpy package should be pip-installed in the python environment

# Mock imports for things not available
# autodoc_mock_imports = ["h5py"]

# -- Project information -----------------------------------------------------

project = 'resqpy'
copyright = '2021, BP'
author = 'BP'

# Version from git tag
# See https://github.com/pypa/setuptools_scm/#usage-from-sphinx
try:
    from importlib import metadata
    release = metadata.version('resqpy')
except Exception:
    release = '0.0.0-version-not-available'

# Take major/minor
version = '.'.join(release.split('.')[:2])

# -- General configuration ---------------------------------------------------

autoclass_content = "class"  # Alternatively, "both" to include init method
# autodoc_default_options = {
#     'members': None,
#     'inherited-members': None,
#     'private-members': None,
#     'show-inheritance': None,
# }
autosummary_generate = True  # Make _autosummary files and include them
autosummary_generate_overwrite = True

napoleon_use_rtype = False  # More legible
autodoc_member_order = 'bysource'
# napoleon_numpy_docstring = False  # Force consistency, leave only Google

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'autoclasstoc',
]

templates_path = ['_templates']

exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']

# -- Options custom autoclasstoc sections ------------------------------------

# See https://autoclasstoc.readthedocs.io/en/latest/advanced_usage.html

from autoclasstoc import PublicMethods


class CommomMethods(PublicMethods):
    key = "common-methods"
    title = "Commonly Used Methods:"

    def predicate(self, name, attr, meta):
        return super().predicate(name, attr, meta) and 'common' in meta


class OtherMethods(PublicMethods):
    key = "other-methods"
    title = "Methods:"

    def predicate(self, name, attr, meta):
        return super().predicate(name, attr, meta) and 'common' not in meta


autoclasstoc_sections = [
    "public-attrs",
    "common-methods",
    "other-methods",
]

# -- Options for HTML output -------------------------------------------------

html_theme = 'sphinx_rtd_theme'

# Override table width restrictions
# See https://rackerlabs.github.io/docs-rackspace/tools/rtd-tables.html
html_static_path = ['_static']
html_context = {
    'css_files': [
        '_static/theme_overrides.css',  # override wide tables in RTD theme
    ],
}



"""Class for RESQML Earth Model Interpretation organizational objects."""

from ._utils import (equivalent_extra_metadata, extract_has_occurred_during, equivalent_chrono_pairs,
                     create_xml_has_occurred_during)

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns
from .organization_feature import OrganizationFeature


class EarthModelInterpretation(BaseResqpy):
    """Class for RESQML Earth Model Interpretation organizational objects."""

    # TODO: add support for StratigraphicColumn reference and other optional references

    resqml_type = 'EarthModelInterpretation'
    valid_domains = ('depth', 'time', 'mixed')

    def __init__(self,
                 parent_model,
                 root_node = None,
                 uuid = None,
                 title = None,
                 organization_feature = None,
                 domain = 'depth',
                 extra_metadata = None):
        """Initialises an earth model interpretation organisational object."""
        self.domain = domain
        self.organization_feature = organization_feature  # InterpretedFeature RESQML field
        self.feature_root = None if self.organization_feature is None else self.organization_feature.root
        self.has_occurred_during = (None, None)
        if (not title) and organization_feature is not None:
            title = organization_feature.feature_name
        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         extra_metadata = extra_metadata,
                         root_node = root_node)

    def _load_from_xml(self):
        self.domain = rqet.find_tag_text(self.root, 'Domain')
        interp_feature_ref_node = rqet.find_tag(self.root, 'InterpretedFeature')
        assert interp_feature_ref_node is not None
        self.feature_root = self.model.referenced_node(interp_feature_ref_node)
        if self.feature_root is not None:
            self.organization_feature = OrganizationFeature(self.model,
                                                            uuid = self.feature_root.attrib['uuid'],
                                                            feature_name = self.model.title_for_root(self.feature_root))
        self.has_occurred_during = extract_has_occurred_during(self.root)

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this interpretation is essentially the same as the other; otherwise False."""
        if other is None or not isinstance(other, EarthModelInterpretation):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        if self.organization_feature is not None:
            if not self.organization_feature.is_equivalent(other.organization_feature):
                return False
        elif other.organization_feature is not None:
            return False
        if self.root is not None and other.root is not None:
            if rqet.citation_title_for_node(self.root) != rqet.citation_title_for_node(other.root):
                return False
        elif self.root is not None or other.root is not None:
            return False
        if check_extra_metadata and not equivalent_extra_metadata(self, other):
            return False
        return self.domain == other.domain and equivalent_chrono_pairs(self.has_occurred_during,
                                                                       other.has_occurred_during)

    def create_xml(self,
                   organization_feature_root = None,
                   add_as_part = True,
                   add_relationships = True,
                   originator = None,
                   title_suffix = None,
                   reuse = True):
        """Creates an earth model interpretation organisational xml node from an earth model interpretation object."""

        # note: related organization feature node should be created first and referenced here

        if not self.title:
            self.title = self.organization_feature.feature_name
        if title_suffix:
            self.title += ' ' + title_suffix

        if reuse and self.try_reuse():
            return self.root
        emi = super().create_xml(add_as_part = False, originator = originator)

        if self.organization_feature is not None:
            of_root = self.organization_feature.root
            if of_root is not None:
                if organization_feature_root is None:
                    organization_feature_root = of_root
                else:
                    assert of_root is organization_feature_root, 'organization feature mismatch'

        assert organization_feature_root is not None, 'interpreted feature not established for model interpretation'

        assert self.domain in self.valid_domains, 'illegal domain value for earth model interpretation'
        dom_node = rqet.SubElement(emi, ns['resqml2'] + 'Domain')
        dom_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
        dom_node.text = self.domain

        self.model.create_ref_node('InterpretedFeature',
                                   self.model.title_for_root(organization_feature_root),
                                   organization_feature_root.attrib['uuid'],
                                   content_type = 'obj_OrganizationFeature',
                                   root = emi)

        create_xml_has_occurred_during(self.model, emi, self.has_occurred_during)

        if add_as_part:
            self.model.add_part('obj_EarthModelInterpretation', self.uuid, emi)
            if add_relationships:
                self.model.create_reciprocal_relationship(emi, 'destinationObject', organization_feature_root,
                                                          'sourceObject')

        return emi



"""surface.py: surface class based on resqml standard."""

version = '4th November 2021'

# RMS and ROXAR are registered trademarks of Roxar Software Solutions AS, an Emerson company
# GOCAD is also a trademark of Emerson

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.olio.triangulation as triangulate
import resqpy.olio.uuid as bu
import resqpy.olio.write_hdf5 as rwh5
import resqpy.olio.xml_et as rqet
from resqpy.olio.xml_namespaces import curly_namespace as ns
from resqpy.olio.zmap_reader import read_mesh
from .base_surface import BaseSurface
from .triangulated_patch import TriangulatedPatch


class Surface(BaseSurface):
    """Class for RESQML triangulated set surfaces."""

    resqml_type = 'TriangulatedSetRepresentation'

    def __init__(self,
                 parent_model,
                 uuid = None,
                 surface_root = None,
                 point_set = None,
                 mesh = None,
                 mesh_file = None,
                 mesh_format = None,
                 tsurf_file = None,
                 quad_triangles = False,
                 title = None,
                 surface_role = 'map',
                 crs_uuid = None,
                 originator = None,
                 extra_metadata = {}):
        """Create an empty Surface object (RESQML TriangulatedSetRepresentation).
        
        Optionally populates from xml, point set or mesh.

        arguments:
           parent_model (model.Model object): the model to which this surface belongs
           uuid (uuid.UUID, optional): if present, the surface is initialised from an existing RESQML object with this uuid
           surface_root (xml tree root node, optional): DEPRECATED: alternative to using uuid
           point_set (PointSet object, optional): if present, the surface is initialised as a Delaunay
              triangulation of the points in the point set; ignored if extracting from xml
           mesh (Mesh object, optional): if present, the surface is initialised as a triangulation of
              the mesh; ignored if extracting from xml or if point_set is present
           mesh_file (string, optional): the path of an ascii file holding a mesh in RMS text or zmap+ format;
              ignored if extracting from xml or point_set or mesh is present
           mesh_format (string, optional): 'rms' or 'zmap'; required if initialising from mesh_file
           tsurf_file (string, optional): the path of an ascii file holding details of triangles and points in GOCAD-Tsurf format;
              ignored if extraction from xml or point_set or mesh is present
           quad_triangles (boolean, default False): if initialising from mesh or mesh_file, each 'square'
              is represented by 2 triangles if quad_triangles is False, 4 triangles if True
           title (string, optional): used as the citation title for the new object, ignored if
              extracting from xml
           surface_role (string, default 'map'): 'map' or 'pick'; ignored if root_node is not None
           crs_uuid (uuid.UUID, optional): if present and not extracting from xml, is set as the crs uuid
              applicable to mesh etc. data
           originator (str, optional): the name of the person creating the object; defaults to login id; ignored
              when initialising from an existing RESQML object
           extra_metadata (dict): items in this dictionary are added as extra metadata; ignored
              when initialising from an existing RESQML object

        returns:
           a newly created surface object

        notes:
           there are 6 ways to initialise a surface object, in order of precendence:
           1. extracting from xml
           2. as a Delaunay triangulation of points in a PointSet
           3. as a simple triangulation of a Mesh object
           4. as a simple triangulation of a mesh in an ascii file
           5. from a GOCAD-TSurf format file
           5. as an empty surface
           if an empty surface is created, 'set_from_...' methods are available to then set for one of:
           - a horizontal plane
           - a single triangle
           - a 'sail' (a triangle wrapped onto a sphere)
           - etc.
           the quad_triangles option is only applied if initialising from a mesh or mesh_file that is fully
           defined (ie. no NaN's)

        :meta common:
        """

        assert surface_role in ['map', 'pick']

        self.surface_role = surface_role
        self.patch_list = []  # ordered list of patches
        self.crs_uuid = crs_uuid
        self.triangles = None  # composite triangles (all patches)
        self.points = None  # composite points (all patches)
        self.boundaries = None  # todo: read up on what this is for and look out for examples
        self.represented_interpretation_root = None
        self.title = title
        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata,
                         root_node = surface_root)
        if self.root is not None:
            pass
        elif point_set is not None:
            self.set_from_point_set(point_set)
        elif mesh is not None:
            self.set_from_mesh_object(mesh, quad_triangles = quad_triangles)
        elif mesh_file and mesh_format:
            self.set_from_mesh_file(mesh_file, mesh_format, quad_triangles = quad_triangles)
        elif tsurf_file is not None:
            self.set_from_tsurf_file(tsurf_file)

    def _load_from_xml(self):
        root_node = self.root
        assert root_node is not None
        self.extract_patches(root_node)
        ref_node = rqet.find_tag(root_node, 'RepresentedInterpretation')
        if ref_node is not None:
            interp_root = self.model.referenced_node(ref_node)
            self.set_represented_interpretation_root(interp_root)

    @property
    def represented_interpretation_uuid(self):
        """Returns the uuid of the represented surface interpretation, or None."""

        return rqet.uuid_for_part_root(self.represented_interpretation_root)

    def set_represented_interpretation_root(self, interp_root):
        """Makes a note of the xml root of the represented interpretation."""

        self.represented_interpretation_root = interp_root

    def extract_patches(self, surface_root):
        """Scan surface root for triangle patches, create TriangulatedPatch objects and build up patch_list."""

        if len(self.patch_list):
            return
        assert surface_root is not None
        paired_list = []
        self.patch_list = []
        for child in surface_root:
            if rqet.stripped_of_prefix(child.tag) != 'TrianglePatch':
                continue
            patch_index = rqet.find_tag_int(child, 'PatchIndex')
            assert patch_index is not None
            triangulated_patch = TriangulatedPatch(self.model, patch_index = patch_index, patch_node = child)
            assert triangulated_patch is not None
            if self.crs_uuid is None:
                self.crs_uuid = triangulated_patch.crs_uuid
            else:
                if not bu.matching_uuids(triangulated_patch.crs_uuid, self.crs_uuid):
                    log.warning('mixed coordinate reference systems in use within a surface')
            paired_list.append((patch_index, triangulated_patch))
        paired_list.sort()
        assert len(paired_list) and paired_list[0][0] == 0 and len(paired_list) == paired_list[-1][0] + 1
        for _, patch in paired_list:
            self.patch_list.append(patch)

    def set_model(self, parent_model):
        """Associate the surface with a resqml model (does not create xml or write hdf5 data)."""

        self.model = parent_model

    def triangles_and_points(self):
        """Returns arrays representing combination of all the patches in the surface.

        Returns:
           Tuple (triangles, points):

           * triangles (int array of shape[:, 3]): integer indices into points array,
             being the nodes of the corners of the triangles
           * points (float array of shape[:, 3]): flat array of xyz points, indexed by triangles

        :meta common:
        """

        if self.triangles is not None:
            return (self.triangles, self.points)
        self.extract_patches(self.root)
        points_offset = 0
        for triangulated_patch in self.patch_list:
            (t, p) = triangulated_patch.triangles_and_points()
            if points_offset == 0:
                self.triangles = t.copy()
                self.points = p.copy()
            else:
                self.triangles = np.concatenate((self.triangles, t.copy() + points_offset))
                self.points = np.concatenate((self.points, p.copy()))
            points_offset += p.shape[0]
        return (self.triangles, self.points)

    def distinct_edges(self):
        """Returns a numpy int array of shape (N, 2) being the ordered node pairs of distinct edges of triangles."""

        triangles, _ = self.triangles_and_points()
        assert triangles is not None
        tri_count = len(triangles)
        all_edges = np.empty((tri_count, 3, 2))
        for i in range(3):
            all_edges[:, i, 0] = triangles[:, i - 1]
            all_edges[:, i, 1] = triangles[:, i]
        return np.unique(np.sort(all_edges.reshape((-1, 2)), axis = 1), axis = 0)

    def set_from_triangles_and_points(self, triangles, points):
        """Populate this (empty) Surface object from an array of triangle corner indices and an array of points."""

        tri_patch = TriangulatedPatch(self.model, patch_index = 0, crs_uuid = self.crs_uuid)
        tri_patch.set_from_triangles_and_points(triangles, points)
        self.patch_list = [tri_patch]
        self.uuid = bu.new_uuid()

    def set_from_point_set(self, point_set, convexity_parameter = 5.0):
        """Populate this (empty) Surface object with a Delaunay triangulation of points in a PointSet object.

        arguments:
           point_set (PointSet): the set of points to be triangulated to form a surface
           convexity_parameter (float, default 5.0): controls how likely the resulting triangulation is to be
              convex; reduce to 1.0 to allow slightly more concavities; increase to 100.0 or more for very little
              chance of even a slight concavity
        """

        p = point_set.full_array_ref()
        log.debug('number of points going into dt: ' + str(len(p)))
        t = triangulate.dt(p[:, :2], container_size_factor = convexity_parameter)
        log.debug('number of triangles: ' + str(len(t)))
        self.crs_uuid = point_set.crs_uuid
        self.set_from_triangles_and_points(t, p)

    def set_from_irregular_mesh(self, mesh_xyz, quad_triangles = False):
        """Populate this (empty) Surface object from an untorn mesh array of shape (N, M, 3).

        arguments:
           mesh_xyz (numpy float array of shape (N, M, 3)): a 2D lattice of points in 3D space
           quad_triangles: (boolean, optional, default False): if True, each quadrangle is represented by
              4 triangles in the surface, with the mean of the 4 corner points used as a common centre node;
              if False (the default), only 2 triangles are used for each quadrangle; note that the 2 triangle
              mode gives a non-unique triangulated result
        """

        mesh_shape = mesh_xyz.shape
        assert len(mesh_shape) == 3 and mesh_shape[2] == 3
        tri_patch = TriangulatedPatch(self.model, patch_index = 0, crs_uuid = self.crs_uuid)
        tri_patch.set_from_irregular_mesh(mesh_xyz, quad_triangles = quad_triangles)
        self.patch_list = [tri_patch]
        self.uuid = bu.new_uuid()

    def set_from_sparse_mesh(self, mesh_xyz):
        """Populate this (empty) Surface object from a mesh array of shape (N, M, 3) with NaNs.

        arguments:
           mesh_xyz (numpy float array of shape (N, M, 3)): a 2D lattice of points in 3D space, with NaNs in z
        """

        mesh_shape = mesh_xyz.shape
        assert len(mesh_shape) == 3 and mesh_shape[2] == 3
        tri_patch = TriangulatedPatch(self.model, patch_index = 0, crs_uuid = self.crs_uuid)
        tri_patch.set_from_sparse_mesh(mesh_xyz)
        self.patch_list = [tri_patch]
        self.uuid = bu.new_uuid()

    def set_from_mesh_object(self, mesh, quad_triangles = False):
        """Populate the (empty) Surface object from a Mesh object."""

        xyz = mesh.full_array_ref()
        if np.any(np.isnan(xyz)):
            self.set_from_sparse_mesh(xyz)
        else:
            self.set_from_irregular_mesh(xyz, quad_triangles = quad_triangles)

    def set_from_torn_mesh(self, mesh_xyz, quad_triangles = False):
        """Populate this (empty) Surface object from a torn mesh array of shape (nj, ni, 2, 2, 3).

        arguments:
           mesh_xyz (numpy float array of shape (nj, ni, 2, 2, 3)): corner points of 2D faces in 3D space
           quad_triangles: (boolean, optional, default False): if True, each quadrangle (face) is represented
              by 4 triangles in the surface, with the mean of the 4 corner points used as a common centre node;
              if False (the default), only 2 triangles are used for each quadrangle; note that the 2 triangle
              mode gives a non-unique triangulated result
        """

        mesh_shape = mesh_xyz.shape
        assert len(mesh_shape) == 5 and mesh_shape[2:] == (2, 2, 3)
        tri_patch = TriangulatedPatch(self.model, patch_index = 0, crs_uuid = self.crs_uuid)
        tri_patch.set_from_torn_mesh(mesh_xyz, quad_triangles = quad_triangles)
        self.patch_list = [tri_patch]
        self.uuid = bu.new_uuid()

    def column_from_triangle_index(self, triangle_index):
        """For surface freshly built from fully defined mesh, returns (j, i) for given triangle index.

        argument:
           triangle_index (int or numpy int array): the triangle index (or array of indices) for which column(s) is/are
           being sought

        returns:
           pair of ints or pair of numpy int arrays: the (j0, i0) indices of the column(s) which the triangle(s) is/are
           part of

        notes:
           this function will only work if the surface has been freshly constructed with data from a mesh without NaNs,
           otherwise (None, None) will be returned;
           the information needed to map from triangle to column is not persistently stored as part of a resqml surface;
           if triangle_index is a numpy int array, a pair of similarly shaped numpy arrays is returned

        :meta common:
        """

        assert len(self.patch_list) == 1
        return self.patch_list[0].column_from_triangle_index(triangle_index)

    def set_to_single_cell_faces_from_corner_points(self, cp, quad_triangles = True):
        """Populates this (empty) surface to represent faces of a cell, from corner points of shape (2, 2, 2, 3)."""

        assert cp.size == 24
        tri_patch = TriangulatedPatch(self.model, patch_index = 0, crs_uuid = self.crs_uuid)
        tri_patch.set_to_cell_faces_from_corner_points(cp, quad_triangles = quad_triangles)
        self.patch_list = [tri_patch]
        self.uuid = bu.new_uuid()

    def set_to_multi_cell_faces_from_corner_points(self, cp, quad_triangles = True):
        """Populates this (empty) surface to represent faces of a set of cells.
        
        From corner points of shape (N, 2, 2, 2, 3).
        """
        assert cp.size % 24 == 0
        cp = cp.reshape((-1, 2, 2, 2, 3))
        self.patch_list = []
        p_index = 0
        for cell_cp in cp:
            tri_patch = TriangulatedPatch(self.model, patch_index = p_index, crs_uuid = self.crs_uuid)
            tri_patch.set_to_cell_faces_from_corner_points(cell_cp, quad_triangles = quad_triangles)
            self.patch_list.append(tri_patch)
            p_index += 1
        self.uuid = bu.new_uuid()

    def cell_axis_and_polarity_from_triangle_index(self, triangle_index):
        """For surface freshly built for cell faces, returns (cell_number, face_axis, polarity).

        arguments:
           triangle_index (int or numpy int array): the triangle index (or array of indices) for which cell face
              information is required

        returns:
           triple int: (cell_number, axis, polarity)

        note:
           if the surface was built for a single cell, the returned cell number will be zero
        """

        triangles_per_face = 4 if self.patch_list[0].quad_triangles else 2
        face_index = triangle_index // triangles_per_face
        cell_number, remainder = divmod(face_index, 6)
        axis, polarity = divmod(remainder, 2)
        return cell_number, axis, polarity

    def set_to_horizontal_plane(self, depth, box_xyz, border = 0.0):
        """Populate this (empty) surface with a patch of two triangles.
        
        Triangles define a flat, horizontal plane at a given depth.

        arguments:
            depth (float): z value to use in all points in the triangulated patch
            box_xyz (float[2, 3]): the min, max values of x, y (&z) giving the area to be covered (z ignored)
            border (float): an optional border width added around the x,y area defined by box_xyz

        :meta common:
        """

        tri_patch = TriangulatedPatch(self.model, patch_index = 0, crs_uuid = self.crs_uuid)
        tri_patch.set_to_horizontal_plane(depth, box_xyz, border = border)
        self.patch_list = [tri_patch]
        self.uuid = bu.new_uuid()

    def set_to_triangle(self, corners):
        """Populate this (empty) surface with a patch of one triangle."""

        tri_patch = TriangulatedPatch(self.model, patch_index = 0, crs_uuid = self.crs_uuid)
        tri_patch.set_to_triangle(corners)
        self.patch_list = [tri_patch]
        self.uuid = bu.new_uuid()

    def set_to_sail(self, n, centre, radius, azimuth, delta_theta):
        """Populate this (empty) surface with a patch representing a triangle wrapped on a sphere."""

        tri_patch = TriangulatedPatch(self.model, patch_index = 0, crs_uuid = self.crs_uuid)
        tri_patch.set_to_sail(n, centre, radius, azimuth, delta_theta)
        self.patch_list = [tri_patch]
        self.uuid = bu.new_uuid()

    def set_from_mesh_file(self, filename, format, quad_triangles = False):
        """Populate this (empty) surface from a zmap or RMS text mesh file."""

        assert format in ['zmap', 'rms', 'roxar']  # 'roxar' is synonymous with 'rms'
        x, y, z = read_mesh(filename, format = format)
        assert x is not None and y is not None or z is not None, 'failed to read surface from zmap file'
        assert x.size == y.size and x.size == z.size, 'non matching array sizes from zmap reader'
        assert x.shape == y.shape and x.shape == z.shape, 'non matching array shapes from zmap reader'

        xyz_mesh = np.stack((x, y, z), axis = -1)
        if np.any(np.isnan(z)):
            self.set_from_sparse_mesh(xyz_mesh)
        else:
            self.set_from_irregular_mesh(xyz_mesh, quad_triangles = quad_triangles)

    def set_from_tsurf_file(self, filename):
        """Populate this (empty) surface from a GOCAD tsurf file."""
        triangles, vertices = [], []
        with open(filename, 'r') as fl:
            lines = fl.readlines()
            for line in lines:
                if "VRTX" in line:
                    vertices.append(line.rstrip().split(" ")[2:])
                elif "TRGL" in line:
                    triangles.append(line.rstrip().split(" ")[1:])
        triangles = np.array(triangles, dtype = int)
        vertices = np.array(vertices, dtype = float)
        self.set_from_triangles_and_points(triangles = triangles, points = vertices)

    def set_from_zmap_file(self, filename, quad_triangles = False):
        """Populate this (empty) surface from a zmap mesh file."""

        self.set_from_mesh_file(filename, 'zmap', quad_triangles = quad_triangles)

    def set_from_roxar_file(self, filename, quad_triangles = False):
        """Populate this (empty) surface from an RMS text mesh file."""

        self.set_from_mesh_file(filename, 'rms', quad_triangles = quad_triangles)

    def set_from_rms_file(self, filename, quad_triangles = False):
        """Populate this (empty) surface from an RMS text mesh file."""

        self.set_from_mesh_file(filename, 'rms', quad_triangles = quad_triangles)

    def vertical_rescale_points(self, ref_depth = None, scaling_factor = 1.0):
        """Modify the z values of points by rescaling.
        
        Stretches the distance from reference depth by scaling factor.
        """
        if scaling_factor == 1.0:
            return
        if ref_depth is None:
            for patch in self.patch_list:
                patch_min = np.min(patch.points[:, 2])
                if ref_depth is None or patch_min < ref_depth:
                    ref_depth = patch_min
        assert ref_depth is not None, 'no z values found for vertical rescaling of surface'
        self.triangles = None  # invalidate any cached triangles & points in surface object
        self.points = None
        for patch in self.patch_list:
            patch.vertical_rescale_points(ref_depth, scaling_factor)

    def write_hdf5(self, file_name = None, mode = 'a'):
        """Create or append to an hdf5 file, writing datasets for the triangulated patches after caching arrays.

        :meta common:
        """

        if self.uuid is None:
            self.uuid = bu.new_uuid()
        # NB: patch arrays must all have been set up prior to calling this function
        h5_reg = rwh5.H5Register(self.model)
        # todo: sort patches by patch index and check sequence
        for triangulated_patch in self.patch_list:
            (t, p) = triangulated_patch.triangles_and_points()
            h5_reg.register_dataset(self.uuid, 'points_patch{}'.format(triangulated_patch.patch_index), p)
            h5_reg.register_dataset(self.uuid, 'triangles_patch{}'.format(triangulated_patch.patch_index), t)
        h5_reg.write(file_name, mode = mode)

    def create_xml(self,
                   ext_uuid = None,
                   add_as_part = True,
                   add_relationships = True,
                   crs_uuid = None,
                   title = None,
                   originator = None):
        """Creates a triangulated surface xml node from this surface object and optionally adds as part of model.

        arguments:
            ext_uuid (uuid.UUID): the uuid of the hdf5 external part holding the surface arrays
            add_as_part (boolean, default True): if True, the newly created xml node is added as a part
                in the model
            add_relationships (boolean, default True): if True, a relationship xml part is created relating the
                new triangulated representation part to the crs part (and optional interpretation part)
            crs_uuid (optional): the uuid of the coordinate reference system applicable to the surface points data;
                if None, the main crs for the model is assumed to apply
            title (string): used as the citation Title text; should be meaningful to a human
            originator (string, optional): the name of the human being who created the triangulated representation part;
                default is to use the login name

        returns:
            the newly created triangulated representation (surface) xml node

        :meta common:
        """

        if ext_uuid is None:
            ext_uuid = self.model.h5_uuid()
        if not self.title:
            self.title = 'surface'

        tri_rep = super().create_xml(add_as_part = False, title = title, originator = originator)

        # todo: if crs_root is None, attempt to derive from surface patch crs uuid (within patch loop, below)
        if crs_uuid is None:
            crs_root = self.model.crs_root  # maverick use of model's default crs
            crs_uuid = rqet.uuid_for_part_root(crs_root)
        else:
            crs_root = self.model.root_for_uuid(crs_uuid)

        if self.represented_interpretation_root is not None:
            interp_root = self.represented_interpretation_root
            interp_uuid = bu.uuid_from_string(interp_root.attrib['uuid'])
            interp_part = self.model.part_for_uuid(interp_uuid)
            interp_title = rqet.find_nested_tags_text(interp_root, ['Citation', 'Title'])
            self.model.create_ref_node('RepresentedInterpretation',
                                       interp_title,
                                       interp_uuid,
                                       content_type = self.model.type_of_part(interp_part),
                                       root = tri_rep)
            if interp_title and not title:
                title = interp_title

        # if not title: title = 'surface'
        # self.model.create_citation(root = tri_rep, title = title, originator = originator)

        role_node = rqet.SubElement(tri_rep, ns['resqml2'] + 'SurfaceRole')
        role_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'SurfaceRole')
        role_node.text = self.surface_role

        for patch in self.patch_list:

            p_node = rqet.SubElement(tri_rep, ns['resqml2'] + 'TrianglePatch')
            p_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'TrianglePatch')
            p_node.text = '\n'

            pi_node = rqet.SubElement(p_node, ns['resqml2'] + 'PatchIndex')
            pi_node.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
            pi_node.text = str(patch.patch_index)

            ct_node = rqet.SubElement(p_node, ns['resqml2'] + 'Count')
            ct_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
            ct_node.text = str(patch.triangle_count)

            cn_node = rqet.SubElement(p_node, ns['resqml2'] + 'NodeCount')
            cn_node.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
            cn_node.text = str(patch.node_count)

            triangles_node = rqet.SubElement(p_node, ns['resqml2'] + 'Triangles')
            triangles_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
            triangles_node.text = '\n'

            # not sure if null value node is needed, not actually used in data
            triangles_null = rqet.SubElement(triangles_node, ns['resqml2'] + 'NullValue')
            triangles_null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
            triangles_null.text = '-1'  # or set to number of points in surface coords?

            triangles_values = rqet.SubElement(triangles_node, ns['resqml2'] + 'Values')
            triangles_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
            triangles_values.text = '\n'

            self.model.create_hdf5_dataset_ref(ext_uuid,
                                               self.uuid,
                                               'triangles_patch{}'.format(patch.patch_index),
                                               root = triangles_values)

            geom = rqet.SubElement(p_node, ns['resqml2'] + 'Geometry')
            geom.set(ns['xsi'] + 'type', ns['resqml2'] + 'PointGeometry')
            geom.text = '\n'

            self.model.create_crs_reference(crs_uuid = crs_uuid, root = geom)

            points_node = rqet.SubElement(geom, ns['resqml2'] + 'Points')
            points_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dHdf5Array')
            points_node.text = '\n'

            coords = rqet.SubElement(points_node, ns['resqml2'] + 'Coordinates')
            coords.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
            coords.text = '\n'

            self.model.create_hdf5_dataset_ref(ext_uuid,
                                               self.uuid,
                                               'points_patch{}'.format(patch.patch_index),
                                               root = coords)

            patch.node = p_node

        if add_as_part:
            self.model.add_part('obj_TriangulatedSetRepresentation', self.uuid, tri_rep)
            if add_relationships:
                # todo: add multiple crs'es (one per patch)?
                self.model.create_reciprocal_relationship(tri_rep, 'destinationObject', crs_root, 'sourceObject')
                if self.represented_interpretation_root is not None:
                    self.model.create_reciprocal_relationship(tri_rep, 'destinationObject',
                                                              self.represented_interpretation_root, 'sourceObject')

                ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False)
                ext_node = self.model.root_for_part(ext_part)
                self.model.create_reciprocal_relationship(tri_rep, 'mlToExternalPartProxy', ext_node,
                                                          'externalPartProxyToMl')

        return tri_rep
